//------------------------------------------------------------------------------
// CERN BE-CO-HT
// White Rabbit Trigger Distribution (WRTD)
// https://ohwr.org/projects/wrtd
//------------------------------------------------------------------------------
//
// unit name:   WrtdDev
//
// description: A SystemVerilog Class to provide an abstraction of a complete
// WRTD device.
//
//------------------------------------------------------------------------------
// Copyright CERN 2019
//------------------------------------------------------------------------------
// Copyright and related rights are licensed under the Solderpad Hardware
// License, Version 2.0 (the "License"); you may not use this file except
// in compliance with the License. You may obtain a copy of the License at
// http://solderpad.org/licenses/SHL-2.0.
// Unless required by applicable law or agreed to in writing, software,
// hardware and materials distributed under this License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
// or implied. See the License for the specific language governing permissions
// and limitations under the License.
//------------------------------------------------------------------------------

`ifndef __WRTD_DEV_INCLUDED
 `define __WRTD_DEV_INCLUDED

 `include "wrtd_definitions.svh"
 `include "wrtd_rule.svh"
 `include "wrtd_alarm.svh"
 `include "mock_turtle_driver.svh"

class WrtdDev;
   protected string name;
   protected MockTurtleDriver    mt;
   protected uint32_t            nbr_cpus;
   protected uint32_t            nbr_devs;
   protected uint32_t            hmq_words[];
   protected wrtd_root           roots[];
   protected WrtdRuleCollection  rules;
   protected WrtdAlarmCollection alarms;

   function new ( CBusAccessor acc, uint64_t base,
		  vIMockTurtleIRQ irq, string name = "" );
      this.name = name;
      this.mt   = new (acc, base, irq, name);
   endfunction // new

   task mdisplay ( string str );
      string tmp;
      if (this.name == "")
	tmp = $sformatf("<%t> %s", $realtime, str);
      else
	tmp = $sformatf("[%s] <%t> %s", this.name, $realtime, str);
      $display (tmp);
   endtask // mdisplay

   task init ( );
      int i, j, n;

      uint32_t val, cpu_ready, init_done, nbr_rules, nbr_alarms;

      uint32_t data[];

      mt.init ( );

      this.nbr_cpus = mt.rom.getCoreCount ( );

      nbr_rules = 0;

      nbr_alarms = 0;

      this.hmq_words = new[this.nbr_cpus];

      this.roots = new[this.nbr_cpus];

      for ( i = 0; i < this.nbr_cpus; i++ )
	begin
	   this.hmq_words[i] =
		`TRTL_CONFIG_ROM_MQ_SIZE_PAYLOAD(mt.rom.getHmqDimensions(i, 0));
	   mt.enable_hmqi_irq ( i, 0, 1 );
	   mt.enable_console_irq ( i, 1 );
	   mt.reset_core ( i, 0 );
	end

      init_done = 0;

      fork
	 begin

	    for ( i = 0; i < this.nbr_cpus; i++ )
	      begin
		 cpu_ready = 0;
		 while ( cpu_ready == 0 )
		   begin
		      val = 0;
		      mt.get_single_cpu_notification ( 0, val );
		      if ( val == TRTL_CPU_NOTIFY_MAIN )
			cpu_ready = 1;
		      # 1us;
		   end
		 msg_get_config ( i, data );
		 this.roots[i].addr = data[0];
		 msg_readw ( i, this.roots[i].addr, `WRTD_ROOT_WORD_SIZE, data );
		 this.roots[i].nbr_rules      = (data[1] & 'h00ff0000) >> 16;
		 this.roots[i].nbr_alarms     = (data[1] & 'h0000ff00) >> 8;
		 this.roots[i].nbr_devs       = (data[1] & 'h000000ff) >> 0;
		 this.roots[i].devs_addr[0:3] = data[4:7];
		 this.roots[i].rules_addr     = data[8];
		 this.roots[i].alarms_addr    = data[9];
		 this.nbr_devs   += this.roots[i].nbr_devs;
		 nbr_rules  += this.roots[i].nbr_rules;
		 nbr_alarms += this.roots[i].nbr_alarms;
	      end

	    this.rules  = new ( nbr_rules, this.name );
	    n = 0;
	    for ( i = 0; i < this.nbr_cpus; i++ )
	      for ( j = 0; j < this.roots[i].nbr_rules; j++ )
		begin
		   this.rules.init_element ( n, i, j );
		   n++;
		end

	    this.alarms = new ( nbr_alarms, this.name );
	    n = 0;
	    for ( i = 0; i < this.nbr_cpus; i++ )
	      for ( j = 0; j < this.roots[i].nbr_alarms; j++ )
		begin
		   this.alarms.init_element ( n, i, j );
		   n++;
		end

	    init_done = 1;
	 end

	 begin
	    while ( init_done == 0 )
	      begin
		 mt.update ( );
		 # 1us;
	      end
	 end
      join

      mdisplay("WRTD ready!");
   endtask // init

   task msg_get_config ( int core, ref uint32_t data[] );
      int i;

      MQueueMsg msg;

      if ( data.size() < `WRTD_CFG_MSG_WORD_SIZE )
	data = new[`WRTD_IO_MSG_WORD_SIZE];

      msg = new ( core, 0 );
      msg.header.flags  = `TRTL_HMQ_HEADER_FLAG_RPC | `TRTL_HMQ_HEADER_FLAG_SYNC;
      msg.header.msg_id = WRTD_ACTION_GET_CONFIG;
      msg.header.len    = 0;
      mt.hmq_send_message ( msg );
      while ( mt.hmq_pending_messages ( msg.core, 0 ) == 0 ) #1us;
      mt.hmq_receive_message ( msg );
      for ( i = 0; i < `WRTD_IO_MSG_WORD_SIZE; i++ )
	data[i] = msg.data[i];

   endtask // msg_get_config

   task msg_readw ( int core, uint32_t addr,
		    uint32_t count, ref uint32_t data[] );
      int i;
      MQueueMsg msg;
      uint32_t tlen, offset;

      if ( data.size() < count )
	data = new[count];

      offset = 0;

      while ( offset <= count )
	begin
	   tlen = this.hmq_words[core] - `WRTD_IO_MSG_WORD_SIZE;
	   if ( tlen > count )
	     tlen = count;
	   msg = new ( core, 0 );
	   msg.header.flags  = `TRTL_HMQ_HEADER_FLAG_RPC | `TRTL_HMQ_HEADER_FLAG_SYNC;
	   msg.header.msg_id = WRTD_ACTION_READW;
	   msg.header.len    = `WRTD_IO_MSG_WORD_SIZE;
	   msg.data          = '{ addr, tlen };
	   mt.hmq_send_message ( msg );
	   while ( mt.hmq_pending_messages ( msg.core, 0 ) == 0 ) #1us;
	   mt.hmq_receive_message ( msg );
	   for ( i = 0; i < tlen; i++ )
	     data[offset+i] = msg.data[i];
	   offset += tlen;
	end

   endtask // msg_readw

   task msg_writew ( int core, uint32_t addr,
		     uint32_t count, uint32_t data[] );
      int i;
      MQueueMsg msg;
      uint32_t tlen, offset;

      offset = 0;

      while ( offset <= count )
	begin
	   tlen = this.hmq_words[core] - `WRTD_IO_MSG_WORD_SIZE;
	   if ( tlen > count )
	     tlen = count;
	   msg = new ( core, 0 );
	   msg.header.flags  = `TRTL_HMQ_HEADER_FLAG_RPC | `TRTL_HMQ_HEADER_FLAG_SYNC;
	   msg.header.msg_id = WRTD_ACTION_WRITEW;
	   msg.header.len    = `WRTD_IO_MSG_WORD_SIZE;
	   msg.data          = '{ addr, tlen };
	   for ( i = 0; i < tlen; i++ )
	     msg.data[i+2] = data[offset+i];
	   mt.hmq_send_message ( msg );
	   while ( mt.hmq_pending_messages ( msg.core, 0 ) == 0 ) #1us;
	   mt.hmq_receive_message ( msg );
	   offset += tlen;
	end

   endtask // msg_writew

   task add_rule ( string rep_cap_id );
      int idx;
      idx = this.rules.add ( rep_cap_id );
      write_rule ( idx );
   endtask // add_rule

   function int find_rule ( string rep_cap_id );
      return this.rules.find ( rep_cap_id );
   endfunction // find_rule

   task remove_rule ( string rep_cap_id );
      int idx;
      idx = this.rules.remove ( rep_cap_id );
      write_rule ( idx );
   endtask // remove_rule

   task set_rule ( string rep_cap_id, WrtdTstamp delay = null,
		   string src = "", string dst = "" );
      int idx;
      idx = this.rules.set ( rep_cap_id, delay, src, dst );
      //apply_rules ();
   endtask // set_rule

/* -----\/----- EXCLUDED -----\/-----
   task apply_rules ( );
      int i, count, src_cpu, dst_cpu;

      for ( i = 0; i < this.rules.collection_size(); i++ )
	begin
	   if ( this.rules.is_free( i ) == 0 )
	     begin
		src_cpu = find_channel ( this.rules.get_src ( i ), WRTD_DIR_INPUT );
		if ( src_cpu < 0 )
		  src_cpu = find_alarm ( this.rules.get_src ( i ) );
		if ( src_cpu < 0 )
		  src_cpu = rx_cpu;
	     end
	end
   endtask // apply_rules
 -----/\----- EXCLUDED -----/\----- */

   task write_rule ( int idx );
      int core, index;
      uint32_t addr;
      uint32_t data[];

      if ( idx < 0 )
	$error ( "cannot write rule with negative index" );

      core  = this.rules.get_core ( idx );
      index = this.rules.get_index (idx );
      addr  = this.roots[core].rules_addr + index * `WRTD_RULE_WORD_SIZE * 4;
      data  = this.rules.data_pack( idx );

      msg_writew ( core, addr, `WRTD_RULE_WORD_SIZE, data );
   endtask // write_rule

   task enable_rule ( int idx );
      this.rules.set_enable ( idx );
      // TODO reconfigure
   endtask // enable_rule

   task add_alarm ( string rep_cap_id );
      int idx;
      idx = this.alarms.add ( rep_cap_id );
      write_alarm ( idx );
   endtask // add_alarm

   function int find_alarm ( string rep_cap_id );
      return this.alarms.find ( rep_cap_id );
   endfunction // find_alarm

   task remove_alarm ( string rep_cap_id );
      int idx;
      idx = this.alarms.remove ( rep_cap_id );
      write_alarm ( idx );
   endtask // remove_alarm

   task write_alarm ( int idx );
      int core, index;
      uint32_t addr;
      uint32_t data[];

      if ( idx < 0 )
	$error ( "cannot write alarm with negative index" );

      core  = this.alarms.get_core ( idx );
      index = this.alarms.get_index (idx );
      addr  = this.roots[core].alarms_addr + index * `WRTD_ALRM_WORD_SIZE * 4;
      data  = this.alarms.data_pack( idx );

      msg_writew ( core, addr, `WRTD_ALRM_WORD_SIZE, data );
   endtask // write_alarm

   task log_read ( );
   endtask // log_read

   task update ( );
      mt.update ( );
      fork
/* -----\/----- EXCLUDED -----\/-----
	 check_log_queue ( 0 );
	 check_log_queue ( 1 );
 -----/\----- EXCLUDED -----/\----- */
      join
   endtask // update

endclass // WrtdDev

`endif //  `ifndef __WRTD_DEV_INCLUDED
