//------------------------------------------------------------------------------
// CERN BE-CO-HT
// White Rabbit Trigger Distribution (WRTD)
// https://ohwr.org/projects/wrtd
//------------------------------------------------------------------------------
//
// unit name:   WrtdRepCap, WrtdRepCapCollection
//
// description: A SystemVerilog Class for a WRTD "repeated capability"
//
//------------------------------------------------------------------------------
// Copyright CERN 2019
//------------------------------------------------------------------------------
// Copyright and related rights are licensed under the Solderpad Hardware
// License, Version 2.0 (the "License"); you may not use this file except
// in compliance with the License. You may obtain a copy of the License at
// http://solderpad.org/licenses/SHL-2.0.
// Unless required by applicable law or agreed to in writing, software,
// hardware and materials distributed under this License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
// or implied. See the License for the specific language governing permissions
// and limitations under the License.
//------------------------------------------------------------------------------

`ifndef __WRTD_REPCAP_INCLUDED
 `define __WRTD_REPCAP_INCLUDED

 `include "wrtd_definitions.svh"

typedef uint32_t repcap_data[];

virtual class WrtdRepCap;
   protected string name;
   protected string rep_cap_id;
   protected int    core;
   protected int    index;
   protected int    enabled;

   function new ( int core, int index, string name = "" );
      this.core  = core;
      this.index = index;
      this.name  = name;
      clear();
   endfunction // new

   function void clear ( );
      this.rep_cap_id = "";
      this.enabled    = 0;
   endfunction // clear

   task mdisplay ( string str );
      string tmp;
      if (this.name == "")
	tmp = $sformatf("<%t> %s", $realtime, str);
      else
	tmp = $sformatf("[%s] <%t> %s", this.name, $realtime, str);
      $display (tmp);
   endtask // mdisplay

   function int get_core ( );
      return this.core;
   endfunction // get_core

   function int get_index ( );
      return this.index;
   endfunction // get_index

   function string get_rep_cap_id ( );
      return this.rep_cap_id;
   endfunction // get_rep_cap_id

   function void set_rep_cap_id ( string id );
      this.rep_cap_id = id;
   endfunction // set_rep_cap_id

   function int match ( string id );
      return ( this.rep_cap_id.compare ( id ) == 0 );
   endfunction // is_equal

   function int is_free ( );
      return ( this.rep_cap_id.len == 0 );
   endfunction // is_free

   function int is_enabled ( );
      return ( this.enabled > 0 );
   endfunction // is_enabled

   function void set_enable ( );
      this.enabled = 1;
   endfunction // set_enable

   function void set_disable ( );
      this.enabled = 0;
   endfunction // set_disable

   pure virtual function repcap_data data_pack ( );

   pure virtual function void data_unpack ( repcap_data data );

endclass //WrtdRepCap

virtual class WrtdRepCapCollection;
   protected string name;
   protected WrtdRepCap collection[];

   function new ( int size, string name );
      this.name       = name;
      this.collection = new[size];
   endfunction // new

   task mdisplay ( string str );
      string tmp;
      if (this.name == "")
	tmp = $sformatf("<%t> %s", $realtime, str);
      else
	tmp = $sformatf("[%s] <%t> %s", this.name, $realtime, str);
      $display (tmp);
   endtask // mdisplay

   function void validate_id ( string rep_cap_id );
      if ( rep_cap_id.len > `WRTD_ID_LEN )
	$error ( "repeated capability name '%s' is too long", rep_cap_id );
      if ( rep_cap_id.len == 0 )
	$error ( "repeated capability name is null" );
   endfunction // validate_id

   function void init_element ( int idx, int core, int rule );
      this.collection[idx] = new ( core, rule, this.name );
   endfunction // init_element

   function int collection_size( );
      return this.collection.size();
   endfunction // collection_size

   function int is_free ( int idx );
      return ( this.collection[idx].is_free() );
   endfunction // is_free

   function int get_core ( int idx );
      return this.collection[idx].get_core();
   endfunction // get_core

   function int get_index ( int idx );
      return this.collection[idx].get_index();
   endfunction // get_index

   function void set_enable ( int idx );
      this.collection[idx].set_enable();
   endfunction // set_enable

   function void set_disable (int idx );
      this.collection[idx].set_disable();
   endfunction // set_disable

   function repcap_data data_pack ( int idx );
      return this.collection[idx].data_pack();
   endfunction // data_pack

   function void data_unpack ( int idx, repcap_data data );
      this.collection[idx].data_unpack ( data );
   endfunction // data_unpack


   function int add ( string rep_cap_id );
      int i, idx;

      validate_id ( rep_cap_id );

      idx = -1;

      for ( i = 0; i < collection_size(); i++ )
	begin

	   if ( this.collection[i].match ( rep_cap_id ) )
	     $error ( "%s repeated capability ID already exists", rep_cap_id );
	   if ( idx == -1 && this.collection[i].is_free() )
	     idx = i;
	end

      if ( idx == -1 )
	$error ( "cannot add %s repeated capability, no space available", rep_cap_id );

      this.collection[idx].set_rep_cap_id ( rep_cap_id );

      return idx;
   endfunction // add

   function int find ( string rep_cap_id );
      int i;

      validate_id ( rep_cap_id );

      for ( i = 0; i < collection_size(); i++ )
	if ( this.collection[i].match ( rep_cap_id ) )
	  return i;

      return -1;
   endfunction // find

   function int remove ( string rep_cap_id );
      int idx;

      idx = find ( rep_cap_id );

      if ( idx == -1 )
	$error ( "%s repeated capability ID cannot be removed because it does not exist", rep_cap_id );

      if ( this.collection[idx].is_enabled() )
	$error ( "%s repeated capability ID cannot be removed because it is enabled", rep_cap_id );

      this.collection[idx].clear();

      return idx;
   endfunction // remove

endclass // WrtdRepCapCollection


`endif //  `ifndef __WRTD_REPCAP_INCLUDED
