//------------------------------------------------------------------------------
// CERN BE-CO-HT
// White Rabbit Trigger Distribution (WRTD)
// https://ohwr.org/projects/wrtd
//------------------------------------------------------------------------------
//
// unit name:   main
//
// description: Testbench for the SPEC150T-based FMC ADC WRTD reference design.
//
//------------------------------------------------------------------------------
// Copyright CERN 2019
//------------------------------------------------------------------------------
// Copyright and related rights are licensed under the Solderpad Hardware
// License, Version 2.0 (the "License"); you may not use this file except
// in compliance with the License. You may obtain a copy of the License at
// http://solderpad.org/licenses/SHL-2.0.
// Unless required by applicable law or agreed to in writing, software,
// hardware and materials distributed under this License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
// or implied. See the License for the specific language governing permissions
// and limitations under the License.
//------------------------------------------------------------------------------

`timescale 1ns/1ps

`include "gn4124_bfm.svh"
`include "wrtd_dev.svh"

`include "fmc_adc_100Ms_csr.v"

`define ADC_CSR_BASE 'h5000

module main;

   IGN4124PCIMaster hostA ();
   //IGN4124PCIMaster hostB ();

   reg       duta_ext_trig, dutb_ext_trig;
   wire[2:0] duta_acq_state, dutb_acq_state;

   dut_env DUTA (hostA, duta_acq_state, a2b_txp, a2b_txn, a2b_rxp, a2b_rxn, duta_ext_trig);
   //dut_env DUTB (hostB, dutb_acq_state, a2b_rxp, a2b_rxn, a2b_txp, a2b_txn, dutb_ext_trig);

   IMockTurtleIRQ MtIrqMonitorA (`MT_ATTACH_IRQ(DUTA.DUT.cmp_mock_turtle));
   //IMockTurtleIRQ MtIrqMonitorB (`MT_ATTACH_IRQ(DUTB.DUT.cmp_mock_turtle));

   CBusAccessor accA, accB;

   WrtdDev devA, devB;

   const uint64_t MT_BASE = 'h0002_0000;

   initial begin

      uint64_t val, expected;

      $timeformat (-6, 3, "us", 10);

      duta_ext_trig <= 1'b0;

      wait (hostA.ready == 1'b1);// && (hostB.ready == 1'b1));

      accA = hostA.get_accessor();
      //accB = hostB.get_accessor();

      accA.set_default_xfer_size(4);
      //accB.set_default_xfer_size(4);

      devA = new (accA, MT_BASE, MtIrqMonitorA, "DUT:A");
      //devB = new (accB, MT_BASE, MtIrqMonitorB, "DUT:B");

      // Config DUTA to trigger on external trigger get 64 samples
      accA.write(`ADC_CSR_BASE + `ADDR_FMC_ADC_100MS_CSR_PRE_SAMPLES,  'h00);
      accA.write(`ADC_CSR_BASE + `ADDR_FMC_ADC_100MS_CSR_POST_SAMPLES, 'h40);
      accA.write(`ADC_CSR_BASE + `ADDR_FMC_ADC_100MS_CSR_SHOTS,        'h01);
      accA.write(`ADC_CSR_BASE + `ADDR_FMC_ADC_100MS_CSR_CH1_GAIN,     'h8000);
      accA.write(`ADC_CSR_BASE + `ADDR_FMC_ADC_100MS_CSR_CH2_GAIN,     'h8000);
      accA.write(`ADC_CSR_BASE + `ADDR_FMC_ADC_100MS_CSR_CH3_GAIN,     'h8000);
      accA.write(`ADC_CSR_BASE + `ADDR_FMC_ADC_100MS_CSR_CH4_GAIN,     'h8000);
      accA.write(`ADC_CSR_BASE + `ADDR_FMC_ADC_100MS_CSR_CH1_SAT,      'h7fff);
      accA.write(`ADC_CSR_BASE + `ADDR_FMC_ADC_100MS_CSR_CH2_SAT,      'h7fff);
      accA.write(`ADC_CSR_BASE + `ADDR_FMC_ADC_100MS_CSR_CH3_SAT,      'h7fff);
      accA.write(`ADC_CSR_BASE + `ADDR_FMC_ADC_100MS_CSR_CH4_SAT,      'h7fff);

      val = (1'b1 << `FMC_ADC_100MS_CSR_TRIG_EN_EXT_OFFSET);
      accA.write(`ADC_CSR_BASE + `ADDR_FMC_ADC_100MS_CSR_TRIG_EN, val);

      expected = 'h39;
      accA.read(`ADC_CSR_BASE + `ADDR_FMC_ADC_100MS_CSR_STA, val);
      if (val != expected)
	$fatal (1, "ADC status error (got 0x%8x, expected 0x%8x).", val, expected);

      $display ("<%t> DUTA ADC configured and armed", $realtime);
      wait (duta_acq_state == 1);

      #500us;

      $display("<%t> START ACQ 1", $realtime);
      accA.write(`ADC_CSR_BASE + `ADDR_FMC_ADC_100MS_CSR_CTL, 'h00000001); // FSM start

      #5us;
      duta_ext_trig <= 1'b1;
      #10ns;
      duta_ext_trig <= 1'b0;

      wait (duta_acq_state == 1);
      $display("<%t> END ACQ 1", $realtime);

      accA.read(`ADC_CSR_BASE + `ADDR_FMC_ADC_100MS_CSR_TRIG_POS, val);
      $display ("<%t> TRIG POSITION %.8x", $realtime, val);

      // DMA transfer
      accA.write('h2008, val); // dma start addr

      accA.write('h200C, 'h00001000); // host addr
      accA.write('h2010, 'h00000000);

      accA.write('h2014, 'h00000100); // len << 2

      accA.write('h2018, 'h00000000); // next
      accA.write('h201C, 'h00000000);

      accA.write('h2020, 'h00000000); // attrib: pcie -> host

      accA.write('h2000, 'h00000001); // xfer start

      wait (DUTA.DUT.dma_irq[0] == 1);
      $display("<%t> END DMA 1", $realtime);

      fork
	 devA.init();
	 //devB.init();
      join

      forever begin
	 devA.update();
	 //devB.update();
	 #1us;
      end

   end

   initial begin
      $display();
      $display("Start of simulation");
      $display("-------------------");
      $display();

      #5000us;

      $display();
      $display("Simulation PASSED");

      $finish;
   end

endmodule // main
