//
// unit name:   ListDriver
//
// description: A SystemVerilog Class to provide an abstraction of a complete
// LIST system.
//
//------------------------------------------------------------------------------
// Copyright CERN 2018
//------------------------------------------------------------------------------
// Copyright and related rights are licensed under the Solderpad Hardware
// License, Version 2.0 (the "License"); you may not use this file except
// in compliance with the License. You may obtain a copy of the License at
// http://solderpad.org/licenses/SHL-2.0.
// Unless required by applicable law or agreed to in writing, software,
// hardware and materials distributed under this License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
// or implied. See the License for the specific language governing permissions
// and limitations under the License.
//------------------------------------------------------------------------------

`ifndef __LIST_DRIVER_INCLUDED
 `define __LIST_DRIVER_INCLUDED

 `include "mock_turtle_driver.svh"

 `define LIST_HMQ_LOG_SLOT 0

typedef struct {
   uint32_t system;
   uint32_t source_port;
   uint32_t trigger;
} list_trig_id;

typedef struct {
   uint64_t seconds;
   uint32_t ticks;
   uint32_t frac;
} wr_timestamp;

typedef enum  {
   NOT_MISSED = -1,
   DEAD_TIME  =  0,
   OVERFLOW   =  1,
   NO_WR      =  2,
   TIMEOUT    =  3
} list_log_miss_reason;

typedef enum  {
   RAW      = (1 << 0),
   SENT     = (1 << 1),
   PROMISC  = (1 << 2),
   FILTERED = (1 << 3),
   EXECUTED = (1 << 4),
   MISSED   = (1 << 5)
} list_log_type;

typedef enum {
   LIST_IN = 2,
   LIST_OUT = 6
} list_log_origin;

class ListLogMsg;
   list_log_origin origin;
   list_log_type ltype;
   uint32_t seq;
   int channel;
   list_trig_id id;
   wr_timestamp ts;
   list_log_miss_reason miss_reason;

   function new ( MQueueMsg msg );
      this.origin         = list_log_origin'(msg.header.msg_id);
      this.ltype          = list_log_type'(msg.data[0]);
      this.seq            = msg.data[1];
      this.channel        = msg.data[2];
      this.id.system      = msg.data[3];
      this.id.source_port = msg.data[4];
      this.id.trigger     = msg.data[5];
      this.ts.seconds     = (msg.data[7] << 32) | msg.data[6];
      this.ts.ticks       = msg.data[8];
      this.ts.frac        = msg.data[9];
      if ( this.ltype == MISSED )
	this.miss_reason = list_log_miss_reason'(msg.data[10]);
      else
	this.miss_reason = NOT_MISSED;
   endfunction // new

   function string tostring ();
      string str;

      str = $sformatf ( {"[%8s] TYPE: %8s, SEQ: %3d, CHANNEL: %2d, ",
			 "TRIG_SYS: %0d, TRIG_PORT: %0d, TRIG_ID: %0d, ",
			 "TS_SEC: %2d, TS_TICKS: %10dns" },
			origin.name(), ltype.name(), seq, channel,
			id.system, id.source_port, id.trigger,
			ts.seconds, ts.ticks * 8);

      if ( ts.frac > 0 )
	str = { str, $sformatf ( ", TS_FRAC: %0d", ts.frac ) };

      if ( miss_reason != NOT_MISSED )
	str = { str, $sformatf ( ", MISS_REASON: %s", miss_reason ) };

      return str;
   endfunction // tostring

endclass // ListLogMsg

class ListDriver;
   protected string name;
   protected int hmq_log_slot;
   protected MockTurtleDriver mt;

   function new ( CBusAccessor acc, uint64_t base,
		  vIMockTurtleIRQ irq, string name = "" );
      this.name = name;
      this.mt   = new (acc, base, irq, name);
   endfunction // new

   task mdisplay ( string str );
      string tmp;
      if (this.name == "")
	tmp = $sformatf("<%t> %s", $realtime, str);
      else
	tmp = $sformatf("[%s] <%t> %s", this.name, $realtime, str);
      $display (tmp);
   endtask // mdisplay

   task init ( string cpu0_fw = "",
	       string cpu1_fw = "");
      mt.init ( );
      mt.enable_console_irq ( 0, 1 );
      mt.enable_console_irq ( 1, 1 );
      mt.enable_hmqi_irq ( 0, 0, 1 );
      mt.enable_hmqi_irq ( 1, 0, 1 );
      mt.reset_core ( 0, 1 );
      mt.reset_core ( 1, 1 );
      if ( cpu0_fw != "" )
	mt.load_firmware ( 0, cpu0_fw, 1'b0 );
      if ( cpu1_fw != "" )
	mt.load_firmware ( 0, cpu1_fw, 1'b0 );
      mt.reset_core ( 0, 0 );
      mt.reset_core ( 1, 0 );
   endtask // init

   task check_log_queue ( int core );
      ListLogMsg log_msg;
      MQueueMsg msg;
      msg = new ( core, `LIST_HMQ_LOG_SLOT );
      while ( mt.hmq_pending_messages ( core, `LIST_HMQ_LOG_SLOT ) )
	begin
	   mt.hmq_receive_message ( msg );
	   log_msg = new ( msg );
	   mdisplay ( log_msg.tostring ( ) );
	end
   endtask // check_log_queue

   task update ( );
      mt.update ( );
      fork
	 check_log_queue ( 0 );
	 check_log_queue ( 1 );
      join
   endtask // update

endclass // ListDriver

`endif //  `ifndef __LIST_DRIVER_INCLUDED
