/*
 * Copyright (C) 2014-2016 CERN (www.cern.ch)
 * Author: Tomasz Wlostowski <tomasz.wlostowski@cern.ch>
 * Author: Federico Vaga <federico.vaga@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __WRTD_COMMON_H
#define __WRTD_COMMON_H

/**
 * WRTD timestamp format
 */
struct wrtd_tstamp {
        // TODO: maybe distinguish between UTC and TAI
	uint32_t seconds;
	uint32_t ns;
	uint32_t frac;
};

/* Length of WRTD ID strings (not null-terminated).  */
#define  WRTD_ID_LEN 16

#define WRTD_ID_NET 0xff

struct wrtd_event {
        /* Time of the event.  */
        struct wrtd_tstamp ts;

        /* Event id.  */
        char id[WRTD_ID_LEN];

        /* Sequence number.  */
        uint32_t seq;

        /* Associated flags.  */
        unsigned char flags;
};

inline static unsigned char wrtd_id_hash(const char *id)
{
        /* This assumes ID is properly padded. */
        return id[0] ^ id[3] ^ id[4] ^ id[7];
}

#define WRTD_DEST_CPU_LOCAL 0xfe
#define WRTD_DEST_CH_NET 0xff

struct wrtd_rule_config {
        /* Id of the rule.  */
        char id[WRTD_ID_LEN];

        char source_id[WRTD_ID_LEN];

        char dest_id[WRTD_ID_LEN];

        /* Device destination.  */
        uint8_t dest_cpu;
        uint8_t dest_ch;

        uint8_t enabled;
        uint8_t send_late;

        uint32_t repeat_count;

        uint32_t delay_ns;
        uint32_t hold_off_ns;

        uint32_t resync_period_ns;
        uint32_t resync_delay_ns;

        /* Next rule id for the same hash; -1 for last.  */
        int32_t hash_chain;
};

struct wrtd_rule_stats {
        uint32_t rx_events;
        struct wrtd_tstamp rx_last;

        uint32_t tx_events;
        struct wrtd_tstamp tx_last;

        /* Latency. */
        uint32_t lat_max_ns;
        uint32_t lat_lo_ns;
        uint32_t lat_hi_ns;
        uint32_t lat_nbr;

        uint32_t miss_holdoff;
        uint32_t miss_late;
        struct wrtd_tstamp miss_last;

        /* Events that occur before this time are discarded.  */
        struct wrtd_tstamp hold_off;

        /* Sequence number over the network.  */
        uint32_t seq;
};

struct wrtd_rule {
        struct wrtd_rule_config conf;
        struct wrtd_rule_stats  stat;
};

struct wrtd_alarm {
        /* Time when to generate the event.  */
        struct wrtd_tstamp setup_time;

        /* Next time and id.  */
        struct wrtd_event event;

        uint32_t enabled;
        int32_t repeat_count;
        uint32_t period_ns;
};

struct wrtd_message {
        unsigned char hw_detect[3]; /* LXI */
        unsigned char domain;       /* 0 */
        unsigned char event_id[WRTD_ID_LEN];
        uint32_t seq;
        uint32_t ts_sec;
        uint32_t ts_ns;
        uint16_t ts_frac;
        uint16_t ts_hi_sec;
        uint8_t flags;
        uint8_t zero[2];
        uint8_t pad[1];
};

enum wrtd_wr_link {
	WR_LINK_OFFLINE = 0,
	WR_LINK_ONLINE,
	WR_LINK_SYNCING,
        WR_LINK_WAIT,
	WR_LINK_SYNCED
};

#define WRTD_VERSION_MAJOR 1
#define WRTD_VERSION_MINOR 4

struct wrtd_root {
        /* Version.  */
        uint8_t ver_major;
        uint8_t ver_minor;
        uint8_t pad0_0;
        uint8_t pad0_1;

        /* Config.  */
        uint8_t nbr_devices;
        uint8_t nbr_alarms;
        uint8_t nbr_rules;
        uint8_t pad1_0;

        /* Status from firmware.  Must not be modified by the user.  */
        enum wrtd_wr_link wr_state;

        /* Status to the firmware.  Not modified by the firmware.  */
        uint8_t log_flags;
        uint8_t freeze_flag;
        uint8_t pad3_0;
        uint8_t pad3_1;

        uint32_t devices_io_addr[4];

        uint32_t rules_addr;
        uint32_t alarms_addr;

        /* Log stats.  */
        /* Number of entries discarded (because flag not set).  */
        uint32_t log_nbr_discarded;
        /* Number of entries missed because rmq/fifo is full.  */
        uint32_t log_nbr_busy;
        /* Number of entries sent.  */
        uint32_t log_nbr_sent;
};

struct wrtd_config_msg {
        /* Always the first one.  */
        uint32_t root_addr;
        uint32_t sync_flag;
        struct wrtd_tstamp now;
};

struct wrtd_io_msg {
        uint32_t addr;
        uint32_t nwords;
        uint32_t data[];
};

#define WRTD_APP_TDC_5CH 0x35D0
#define WRTD_APP_FD_4CH  0x35E0
#define WRTD_APP_ADCIO_5CH 0x35B0

enum wrtd_trtl_actions {
        /* Always the first one, to get root and version.  */
        WRTD_ACTION_GET_CONFIG,
        WRTD_ACTION_READW,
        WRTD_ACTION_WRITEW
};

#define WRTD_ACTION_LOG 0x20

enum wrtd_log_msg_type {
        WRTD_LOG_MSG_EV_GENERATED,
        WRTD_LOG_MSG_EV_CONSUMED,
        WRTD_LOG_MSG_EV_DISCARDED,
        WRTD_LOG_MSG_EV_NETWORK,
        WRTD_LOG_MSG_LINK
};

enum wrtd_log_reason_type {
        /* For GENERATED: */
        WRTD_LOG_GENERATED_ALARM = 1,
        WRTD_LOG_GENERATED_DEVICE = 8,
        WRTD_LOG_GENERATED_DEVICE_0 = WRTD_LOG_GENERATED_DEVICE + 0,
        WRTD_LOG_GENERATED_DEVICE_1 = WRTD_LOG_GENERATED_DEVICE + 8,
        WRTD_LOG_GENERATED_DEVICE_2 = WRTD_LOG_GENERATED_DEVICE + 16,
        WRTD_LOG_GENERATED_DEVICE_3 = WRTD_LOG_GENERATED_DEVICE + 24,

        /* For CONSUMED: */
        WRTD_LOG_CONSUMED_START = 1,
        WRTD_LOG_CONSUMED_DONE,

        /* For DISCARDED: */
        /* Not synchronized.  */
        WRTD_LOG_DISCARD_NO_SYNC = 1,
        /* Timeout: event after time.  */
        WRTD_LOG_DISCARD_TIMEOUT,
        /* No space in a queue.  */
        WRTD_LOG_DISCARD_OVERFLOW,

        /* For NETWORK: */
        WRTD_LOG_NETWORK_TX = 1,
        WRTD_LOG_NETWORK_RX

};

/**
 * Log event descriptor
 */
struct wrtd_log_entry {
	uint32_t type;
        uint32_t reason;
        struct wrtd_event event;
        struct wrtd_tstamp ts;
};


/* Remote message queue for host and log.  */
#define WRTD_HMQ 0

#endif
