#include <stdarg.h>
#include <stdlib.h>
#include <errno.h>
#include "libmockturtle.h"
#include "libwrtd.h"
#include "libwrtd-private.h"

/* Number of existing alarms.  */
enum wrtd_status wrtd_attr_get_alarm_count(struct wrtd_dev *wrtd,
                                           int32_t *value)
{
        enum wrtd_status status;
        unsigned i;
        int32_t res;

        status = wrtd_fill_alarms(wrtd);
        WRTD_RETURN_IF_ERROR(status);

        res = 0;
        for (i = 0; i < wrtd->nbr_alarms; i++)
                if (!wrtd_id_null(wrtd->alarms[i].alarm.event.id))
                        res++;
        *value = res;
        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_alarm_check_disabled(struct wrtd_dev *wrtd, unsigned idx)
{
        if (wrtd->alarms[idx].alarm.enabled)
                return wrtd_return_error(wrtd, WRTD_ERROR_RESOURCE_ACTIVE,
                                         "alarm is enabled");
        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_set_alarm_time(struct wrtd_dev *wrtd,
                                          const char *rep_cap_id,
                                          const struct wrtd_tstamp *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_alarm(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_alarm_check_disabled(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        wrtd->alarms[idx].alarm.event.ts = *value;

        status = wrtd_write_alarm(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_set_alarm_period(struct wrtd_dev *wrtd,
                                            const char *rep_cap_id,
                                            const struct wrtd_tstamp *value)
{
        enum wrtd_status status;
        unsigned int idx;

        if (value->seconds != 0)
                return wrtd_return_error(wrtd, WRTD_ERROR_INVALID_VALUE,
                                         "alarm period is too large (>1s)");

        status = wrtd_find_alarm(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_alarm_check_disabled(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        wrtd->alarms[idx].alarm.period_ns = value->ns;

        status = wrtd_write_alarm(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_set_alarm_setup_time(struct wrtd_dev *wrtd,
                                                const char *rep_cap_id,
                                                const struct wrtd_tstamp *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_alarm(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_alarm_check_disabled(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        wrtd->alarms[idx].alarm.setup_time = *value;

        status = wrtd_write_alarm(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_set_alarm_enable(struct wrtd_dev *wrtd,
                                            const char *rep_cap_id,
                                            bool value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_alarm(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        wrtd->alarms[idx].alarm.enabled = value;

        /* FIXME: may modify the rule mapping (eg: alarm is placed on cpu,
           and need to trigger an rule initially placed on a different cpu).
        */
        status = wrtd_write_alarm(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_alarm_enable(struct wrtd_dev *wrtd,
                                            const char *rep_cap_id,
                                            bool *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_alarm(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        *value = wrtd->alarms[idx].alarm.enabled != 0;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_set_alarm_repeat_count(struct wrtd_dev *wrtd,
                                                  const char *rep_cap_id,
                                                  int32_t count)
{
        enum wrtd_status status;
        unsigned int idx;

        if (count < 0)
                return wrtd_return_error(wrtd, WRTD_ERROR_INVALID_VALUE,
                                         "alarm repeat count < 0");

        status = wrtd_find_alarm(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_alarm_check_disabled(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        wrtd->alarms[idx].alarm.repeat_count = count == 0 ? 0 : count + 1;

        status = wrtd_write_alarm(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_set_log_enable(struct wrtd_dev *wrtd,
                                          bool value)
{
        enum wrtd_status status;
        unsigned int cpu;

        status = wrtd_fill_roots(wrtd);
        WRTD_RETURN_IF_ERROR(status);

        for (cpu = 0; cpu < wrtd->nbr_cpus; cpu++) {
                wrtd->roots[cpu].log_flags = value ? 0xff : 0;
                status = wrtd_write_root_flags(wrtd, cpu);
                WRTD_RETURN_IF_ERROR(status);
        }

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_rule_check_disabled(struct wrtd_dev *wrtd, unsigned idx)
{
        if (wrtd->rules[idx].rule.conf.enabled)
                return wrtd_return_error(wrtd, WRTD_ERROR_RESOURCE_ACTIVE,
                                         "rule is enabled");
        return WRTD_SUCCESS;
}


enum wrtd_status wrtd_find_rule_check(struct wrtd_dev *wrtd,
                                      const char *rep_cap_id,
                                      unsigned *idx)
{
        enum wrtd_status status;

        status = wrtd_find_rule(wrtd, rep_cap_id, idx);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_rule_check_disabled(wrtd, *idx);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_rule_count(struct wrtd_dev *wrtd,
                                           unsigned *value)
{
        enum wrtd_status status;
        unsigned i;
        unsigned res;

        status = wrtd_fill_rules(wrtd);
        WRTD_RETURN_IF_ERROR(status);

        res = 0;
        for (i = 0; i < wrtd->nbr_rules; i++)
                if (!wrtd_id_null(wrtd->rules[i].rule.conf.id))
                        res++;
        *value = res;
        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_set_rule_source(struct wrtd_dev *wrtd,
                                           const char *rep_cap_id,
                                           const char *src)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_validate_id(wrtd, src);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_find_rule_check(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        wrtd_id_copy(wrtd->rules[idx].rule.conf.source_id, src);

        status = wrtd_write_rule(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_set_rule_destination(struct wrtd_dev *wrtd,
                                                const char *rep_cap_id,
                                                const char *dst)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_validate_id(wrtd, dst);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_find_rule_check(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        wrtd_id_copy(wrtd->rules[idx].rule.conf.dest_id, dst);

        status = wrtd_write_rule(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_rule_destination(struct wrtd_dev *wrtd,
                                                const char *rep_cap_id,
                                                int32_t buf_size,
                                                char *dst)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_id_copy_buf(wrtd, dst, buf_size,
                                  wrtd->rules[idx].rule.conf.dest_id);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_rule_source(struct wrtd_dev *wrtd,
                                           const char *rep_cap_id,
                                           int32_t buf_size,
                                           char *dst)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_id_copy_buf(wrtd, dst, buf_size,
                                  wrtd->rules[idx].rule.conf.source_id);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_set_rule_delay(struct wrtd_dev *wrtd,
                                          const char *rep_cap_id,
                                          const struct wrtd_tstamp *value)
{
        enum wrtd_status status;
        unsigned int idx;

        if (value->seconds != 0)
                return wrtd_return_error(wrtd, WRTD_ERROR_INVALID_VALUE,
                                         "rule delay is too large (>1s)");

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_rule_check_disabled(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        wrtd->rules[idx].rule.conf.delay_ns = value->ns;

        status = wrtd_write_rule(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_get_rule_enable(struct wrtd_dev *wrtd,
                                           const char *rep_cap_id,
                                           bool *value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        *value = wrtd->rules[idx].rule.conf.enabled != 0;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_set_rule_enable(struct wrtd_dev *wrtd,
                                           const char *rep_cap_id,
                                           bool value)
{
        enum wrtd_status status;
        unsigned int idx;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        if (wrtd->rules[idx].rule.conf.enabled != value) {
                wrtd->rules[idx].rule.conf.enabled = value;

                status = wrtd_reconfigure(wrtd);
                WRTD_RETURN_IF_ERROR(status);
        }

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_attr_global(struct wrtd_dev *wrtd,
                                  const char *rep_cap_id)
{
        if (!wrtd_id_eq(rep_cap_id, WRTD_GLOBAL_REP_CAP_ID))
                return wrtd_return_error(wrtd, WRTD_ERROR_ATTR_INVALID_REP_CAP,
                                         "global repeated capability required");
        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_get_attr_int32(struct wrtd_dev *wrtd,
                                     const char *rep_cap_id,
                                     enum wrtd_attr id,
                                     int32_t *value)
{
        enum wrtd_status status;

        status = wrtd_validate_id(wrtd, rep_cap_id);
        WRTD_RETURN_IF_ERROR(status);

        switch(id) {
        case WRTD_MAJOR_VERSION:
                status = wrtd_attr_global(wrtd, rep_cap_id);
                WRTD_RETURN_IF_ERROR(status);
                *value = WRTD_VERSION_MAJOR;
                return WRTD_SUCCESS;
        case WRTD_MINOR_VERSION:
                status = wrtd_attr_global(wrtd, rep_cap_id);
                WRTD_RETURN_IF_ERROR(status);
                *value = WRTD_VERSION_MINOR;
                return WRTD_SUCCESS;

        case WRTD_ATTR_ALARM_COUNT:
                status = wrtd_attr_global(wrtd, rep_cap_id);
                WRTD_RETURN_IF_ERROR(status);
                status = wrtd_attr_get_alarm_count(wrtd, value);
                WRTD_RETURN_IF_ERROR(status);
                return WRTD_SUCCESS;

        case WRTD_ATTR_RULE_COUNT:
        {
                unsigned val;
                status = wrtd_attr_global(wrtd, rep_cap_id);
                WRTD_RETURN_IF_ERROR(status);
                status = wrtd_attr_get_rule_count(wrtd, &val);
                WRTD_RETURN_IF_ERROR(status);
                *value = val;
                return WRTD_SUCCESS;
        }
        default:
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_INVALID_ATTRIBUTE,
                         "int32 attribute %u is not known", id);
        }
}

enum wrtd_status wrtd_set_attr_tstamp(struct wrtd_dev *wrtd,
                                      const char *rep_cap_id,
                                      enum wrtd_attr id,
                                      const struct wrtd_tstamp *value)
{
        enum wrtd_status status;

        status = wrtd_validate_id(wrtd, rep_cap_id);
        WRTD_RETURN_IF_ERROR(status);

        switch(id) {
        case WRTD_ATTR_ALARM_TIME:
                return wrtd_attr_set_alarm_time(wrtd, rep_cap_id, value);
        case WRTD_ATTR_ALARM_SETUP_TIME:
                return wrtd_attr_set_alarm_setup_time(wrtd, rep_cap_id, value);
        case WRTD_ATTR_ALARM_PERIOD:
                return wrtd_attr_set_alarm_period(wrtd, rep_cap_id, value);
        case WRTD_ATTR_RULE_DELAY:
                return wrtd_attr_set_rule_delay(wrtd, rep_cap_id, value);
        default:
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_INVALID_ATTRIBUTE,
                         "tstamp attribute %u is not known", id);
        }
}

enum wrtd_status wrtd_set_attr_bool(struct wrtd_dev *wrtd,
                                    const char *rep_cap_id,
                                    enum wrtd_attr id,
                                    bool value)
{
        enum wrtd_status status;

        status = wrtd_validate_id(wrtd, rep_cap_id);
        WRTD_RETURN_IF_ERROR(status);

        switch(id) {
        case WRTD_ATTR_EVENT_LOG_ENABLED:
                status = wrtd_attr_global(wrtd, rep_cap_id);
                WRTD_RETURN_IF_ERROR(status);
                return wrtd_attr_set_log_enable(wrtd, value);
        case WRTD_ATTR_ALARM_ENABLED:
                return wrtd_attr_set_alarm_enable(wrtd, rep_cap_id, value);
        case WRTD_ATTR_RULE_ENABLED:
                return wrtd_attr_set_rule_enable(wrtd, rep_cap_id, value);
        default:
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_INVALID_ATTRIBUTE,
                         "bool attribute %u is not known", id);
        }
}

enum wrtd_status wrtd_get_attr_bool(struct wrtd_dev *wrtd,
                                    const char *rep_cap_id,
                                    enum wrtd_attr id,
                                    bool *value)
{
       enum wrtd_status status;

        status = wrtd_validate_id(wrtd, rep_cap_id);
        WRTD_RETURN_IF_ERROR(status);

        switch(id) {
        case WRTD_ATTR_ALARM_ENABLED:
                return wrtd_attr_get_alarm_enable(wrtd, rep_cap_id, value);
        case WRTD_ATTR_RULE_ENABLED:
                return wrtd_attr_get_rule_enable(wrtd, rep_cap_id, value);
        default:
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_INVALID_ATTRIBUTE,
                         "bool attribute %u is not known", id);
        }
}

enum wrtd_status wrtd_set_attr_int32(struct wrtd_dev *wrtd,
                                     const char *rep_cap_id,
                                     enum wrtd_attr id,
                                     int32_t value)
{
        enum wrtd_status status;

        status = wrtd_validate_id(wrtd, rep_cap_id);
        WRTD_RETURN_IF_ERROR(status);

        switch(id) {
        case WRTD_ATTR_ALARM_REPEAT_COUNT:
                return wrtd_attr_set_alarm_repeat_count
                        (wrtd, rep_cap_id, value);
        default:
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_INVALID_ATTRIBUTE,
                         "int32 attribute %u is not known", id);
        }
}

enum wrtd_status wrtd_set_attr_string(struct wrtd_dev *wrtd,
                                      const char *rep_cap_id,
                                      enum wrtd_attr id,
                                      const char *value)
{
        enum wrtd_status status;

        status = wrtd_validate_id(wrtd, rep_cap_id);
        WRTD_RETURN_IF_ERROR(status);

        switch(id) {
        case WRTD_ATTR_RULE_SOURCE:
                return wrtd_attr_set_rule_source
                        (wrtd, rep_cap_id, value);
        case WRTD_ATTR_RULE_DESTINATION:
                return wrtd_attr_set_rule_destination
                        (wrtd, rep_cap_id, value);
        default:
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_INVALID_ATTRIBUTE,
                         "string set attribute %u is not known", id);
        }
}

enum wrtd_status wrtd_get_attr_string(struct wrtd_dev *wrtd,
                                      const char *rep_cap_id,
                                      enum wrtd_attr id,
                                      int32_t value_buf_size,
                                      char *value)
{
        enum wrtd_status status;

        status = wrtd_validate_id(wrtd, rep_cap_id);
        WRTD_RETURN_IF_ERROR(status);

        switch(id) {
        case WRTD_ATTR_RULE_SOURCE:
                return wrtd_attr_get_rule_source
                        (wrtd, rep_cap_id, value_buf_size, value);
        case WRTD_ATTR_RULE_DESTINATION:
                return wrtd_attr_get_rule_destination
                        (wrtd, rep_cap_id, value_buf_size, value);
        default:
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_INVALID_ATTRIBUTE,
                         "string get attribute %u is not known", id);
        }
}
