#include <stdarg.h>
#include <stddef.h>
#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include "libmockturtle.h"
#include "libwrtd.h"
#include "libwrtd-private.h"

#define WRTD_DEFAULT_TIMEOUT 1000

static void init (void)
{
        static int initialized;

        if (initialized)
                return;
        initialized = 1;
        trtl_init();
}

enum wrtd_status wrtd_return_error(struct wrtd_dev *dev,
                                   enum wrtd_status error_code,
                                   const char *format, ...)
{
        va_list args;

        va_start(args, format);
        dev->err = error_code;
        vsnprintf(dev->error_msg, sizeof(dev->error_msg), format, args);
        va_end(args);

        return error_code;
}

const char *wrtd_get_error_msg(struct wrtd_dev *dev)
{
        return dev->error_msg;
}

enum wrtd_status wrtd_init(const char *resource_name,
                           bool reset,
                           const char *options_str,
                           struct wrtd_dev **dev)
{
        struct trtl_dev *trtl;

        /* In case of error...  */
        *dev = NULL;

        /* Initialize (only once).
           FIXME: check errors ? (but there are no errors).  */
        init ();

        /* Resource is MTxxx where xxx is the mock-turtle device id.  */
        if (resource_name[0] == 'M' && resource_name[1] == 'T') {
                unsigned long dev_id;
                char *eptr;
                dev_id = strtoul(resource_name + 2, &eptr, 0);
                if (*eptr != 0) {
                        /* Invalid characters.  */
                        return WRTD_ERROR_RESOURCE_UNKNOWN;
                }
                trtl = trtl_open_by_id(dev_id);
        }
        else if (strncmp(resource_name, "trtl-", 5) == 0) {
                trtl = trtl_open(resource_name);
        }
        else {
                return WRTD_ERROR_RESOURCE_UNKNOWN;
        }
        if (trtl == NULL)
                return WRTD_ERROR_RESOURCE_UNKNOWN;

        struct wrtd_dev *res;

        res = malloc(sizeof(struct wrtd_dev));
        if (res == NULL)
                return WRTD_ERROR_OUT_OF_MEMORY;
        memset(res, 0, sizeof(*res));

 	res->trtl = trtl;

        *dev = res;
        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_close(struct wrtd_dev *dev)
{
        trtl_close(dev->trtl);

        free(dev->alarms);
        //free(dev->rules);
        free(dev);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_reset(struct wrtd_dev *dev)
{
        return WRTD_SUCCESS;
}


enum wrtd_status wrtd_get_firmware_id(struct wrtd_dev *dev,
                                      unsigned *nbr_ids,
                                      uint32_t *ids)
{
        const struct wrtd_mt_config *cfg;
        enum wrtd_status status;
        int i;

        /* Find config.  */
        status = wrtd_find_config(dev, &cfg);
        WRTD_RETURN_IF_ERROR(status);

        /* Find firmwares.  */
        *nbr_ids = cfg->nbr_cpus;
        for (i = 0; i < cfg->nbr_cpus; i++) {
                ids[i] = cfg->cpus[i].fw_id;
        }
        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_load_firmware(struct wrtd_dev *dev, unsigned verbose,
                                    unsigned int nbr_firmwares,
                                    struct wrtd_firmware *fw)
{
	struct trtl_fw_version version;
        const struct wrtd_mt_config *cfg;
        enum wrtd_status status;
	int err;
        int i, j;
        struct wrtd_firmware *map[WRTD_MAX_CPUS];

        /* Find config.  */
        status = wrtd_find_config(dev, &cfg);
        WRTD_RETURN_IF_ERROR(status);

        /* Find firmwares.  */
        for (i = 0; i < cfg->nbr_cpus; i++) {
                map[i] = NULL;
                for (j = 0; j < nbr_firmwares; j++)
                        if (cfg->cpus[i].fw_id == fw[j].id) {
                                map[i] = &fw[j];
                                break;
                        }
                if (map[i] == NULL) {
                        /* Missing firmware.  */
                        return wrtd_return_error
                                (dev, WRTD_ERROR_RESOURCE_UNKNOWN,
                                 "no firmware for app 0x%x",
                                 cfg->cpus[i].fw_id);
                }
        }

        /* Load firmwares.  */
        for (i = 0; i < cfg->nbr_cpus; i++) {
                if (verbose)
                        printf("loading cpu %d with %s\n", i, map[i]->path);
                err = trtl_cpu_load_application_file (dev->trtl,
                                                      i, map[i]->path);
                if (err)
                        return wrtd_return_error
                                (dev, WRTD_ERROR_UNEXPECTED_RESPONSE,
                                 "cannot load application");

                if (verbose)
                        printf("start cpu %d\n", i);
                err = trtl_cpu_enable(dev->trtl, i);
                if (err)
                        return wrtd_return_error
                                (dev, WRTD_ERROR_UNEXPECTED_RESPONSE,
                                 "cannot start cpu");
        }

        /* Check CPUs.  */
        for (i = 0; i < cfg->nbr_cpus; i++) {
                if (verbose)
                        printf("waiting cpu %d to be ready\n", i);

                /* Do a ping, but wait for 5sec for the FW to be ready.  */
                err = trtl_fw_ping_timeout(dev->trtl, i, 0, 5000);
                if (err)
                        return wrtd_return_error
                                (dev, WRTD_ERROR_UNEXPECTED_RESPONSE,
                                 "ping failed on cpu %d", i);

                err = trtl_fw_version(dev->trtl, i, 0, &version);
                if (err)
                        return wrtd_return_error
                                (dev, WRTD_ERROR_UNEXPECTED_RESPONSE,
                                 "cannot get version on cpu %d", i);

                if (version.rt_id != map[i]->id)
                        return wrtd_return_error
                                (dev, WRTD_ERROR_UNEXPECTED_RESPONSE,
                                 "firmware id mismatch on cpu %d", i);
        }

        return WRTD_SUCCESS;
}


enum wrtd_status wrtd_msg_get_config(struct wrtd_dev *wrtd, unsigned cpu,
                                     struct wrtd_config_msg *res)
{
	struct trtl_msg msg;
	int err;

	memset(&msg.hdr, 0, sizeof(struct trtl_hmq_header));
	msg.hdr.flags = TRTL_HMQ_HEADER_FLAG_RPC;
	msg.hdr.msg_id = WRTD_ACTION_GET_CONFIG;
        msg.hdr.len = 0;

	err = trtl_msg_sync(wrtd->trtl, cpu, 0, &msg, &msg,
                            WRTD_DEFAULT_TIMEOUT);

	if (err)
                return wrtd_return_error(wrtd, WRTD_ERROR_UNEXPECTED_RESPONSE,
                                         "sync msg error(get_config): %s",
                                         trtl_strerror(errno));

	if (msg.hdr.msg_id != WRTD_ACTION_GET_CONFIG
            || msg.hdr.len != 5)
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_UNEXPECTED_RESPONSE,
                         "sync msg (get_config): bad id 0x%x/len %d",
                         msg.hdr.msg_id, msg.hdr.len);


        res->root_addr = msg.data[0];
        res->sync_flag = msg.data[1];
        res->now.seconds = msg.data[2];
        res->now.ns = msg.data[3];
        res->now.frac = msg.data[4];

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_get_sys_time(struct wrtd_dev *wrtd,
                                   struct wrtd_tstamp *time)
{
        struct wrtd_config_msg msg;
        enum wrtd_status status;

        /* Always use CPU #0 (it does exist!).  */
        status = wrtd_msg_get_config(wrtd, 0, &msg);
        WRTD_RETURN_IF_ERROR(status);

        *time = msg.now;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_fill_config(struct wrtd_dev *wrtd)
{
        int i;
        const struct trtl_config_rom *cfgrom;
        struct wrtd_config_msg msg;
        enum wrtd_status status;

        /* Need to be fetched only once.  */
        if (wrtd->config != NULL)
                return WRTD_SUCCESS;

        status = wrtd_find_config(wrtd, &wrtd->config);
        WRTD_RETURN_IF_ERROR(status);

        cfgrom = trtl_config_get(wrtd->trtl);

        /* Set HMQ words.  */
        for (i = 0; i < cfgrom->n_cpu; i++)
                if (cfgrom->n_hmq[i] < 0)
                        wrtd->hmq_words[i] = TRTL_CONFIG_ROM_MQ_SIZE_PAYLOAD
                                (cfgrom->hmq[i][0].sizes);

        if (cfgrom->n_cpu > WRTD_MAX_CPUS)
                return wrtd_return_error(wrtd, WRTD_ERROR_INVALID_VALUE,
                                         "incorrect number of cpus");

        for (i = 0; i < cfgrom->n_cpu; i++) {
                status = wrtd_msg_get_config(wrtd, i, &msg);
                WRTD_RETURN_IF_ERROR(status);
                wrtd->root_addr[i] = msg.root_addr;
                /* FIXME: check firmware version ?  */
        }

        wrtd->nbr_cpus = cfgrom->n_cpu;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_fill_roots(struct wrtd_dev *wrtd)
{
        enum wrtd_status status;
        unsigned cpu;

        if (wrtd->roots[0].ver_major != 0)
                return WRTD_SUCCESS;

        status = wrtd_fill_config(wrtd);
        WRTD_RETURN_IF_ERROR(status);

        for (cpu = 0; cpu < wrtd->nbr_cpus; cpu++) {
                status = wrtd_msg_readw(wrtd, cpu,
                                        wrtd->root_addr[cpu],
                                        sizeof(struct wrtd_root) / 4,
                                        (uint32_t*)&wrtd->roots[cpu]);
                if (status != WRTD_SUCCESS)
                        return status;
                if (wrtd->roots[cpu].ver_major != WRTD_VERSION_MAJOR) {
                        status = wrtd_return_error
                                (wrtd, WRTD_ERROR_UNEXPECTED_RESPONSE,
                                 "incorrect major version (%d instead of %d)",
                                 wrtd->roots[cpu].ver_major,
                                 WRTD_VERSION_MAJOR);
                        /* Invalidate root.  */
                        wrtd->roots[cpu].ver_major = 0;
                        return status;
                }
                if (wrtd->roots[cpu].ver_minor != WRTD_VERSION_MINOR) {
                        /* Invalidate root.  */
                        wrtd->roots[cpu].ver_major = 0;
                        return wrtd_return_error
                                (wrtd, WRTD_ERROR_UNEXPECTED_RESPONSE,
                                 "incorrect major version (%d instead of %d)",
                                 wrtd->roots[cpu].ver_minor,
                                 WRTD_VERSION_MINOR);
                }
        }
        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_fill_alarms(struct wrtd_dev *wrtd)
{
        enum wrtd_status status;
        unsigned cpu;
        unsigned n;

        if (wrtd->alarms != NULL)
                return WRTD_SUCCESS;

        status = wrtd_fill_roots(wrtd);
        WRTD_RETURN_IF_ERROR(status);

        /* Count number of alarms.  */
        for (cpu = 0; cpu < wrtd->nbr_cpus; cpu++)
                wrtd->nbr_alarms += wrtd->roots[cpu].nbr_alarms;

        /* Allocates.  */
        wrtd->alarms = (struct wrtd_lib_alarm *)
                malloc(wrtd->nbr_alarms * sizeof(struct wrtd_lib_alarm));
        if (wrtd->alarms == NULL)
                return wrtd_return_error(wrtd, WRTD_ERROR_OUT_OF_MEMORY,
                                         "cannot allocate %u alarms",
                                         wrtd->nbr_alarms);
        n = 0;
        for (cpu = 0; cpu < wrtd->nbr_cpus; cpu++) {
                unsigned idx;
                for (idx = 0; idx < wrtd->roots[cpu].nbr_alarms; idx++) {
                        status = wrtd_msg_readw
                                (wrtd, cpu,
                                 (wrtd->roots[cpu].alarms_addr
                                  + idx * sizeof(struct wrtd_alarm)),
                                 sizeof(struct wrtd_alarm) / 4,
                                 (uint32_t *)&wrtd->alarms[n].alarm);
                        if (status != WRTD_SUCCESS) {
                                free(wrtd->alarms);
                                wrtd->alarms = NULL;
                                return status;
                        }
                        wrtd->alarms[n].cpu = cpu;
                        wrtd->alarms[n].idx = idx;
                        n++;
                }
        }

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_write_alarm(struct wrtd_dev *wrtd, unsigned idx)
{
        enum wrtd_status status;
        struct wrtd_lib_alarm *al = &wrtd->alarms[idx];

        status = wrtd_msg_writew
                (wrtd, al->cpu,
                 (wrtd->roots[al->cpu].alarms_addr
                  + al->idx * sizeof(struct wrtd_alarm)),
                 sizeof(struct wrtd_alarm) / 4,
                 (uint32_t *)&al->alarm);

        return status;
}

enum wrtd_status wrtd_msg_readw(struct wrtd_dev *wrtd, unsigned cpu,
                                uint32_t addr, uint32_t count,
                                uint32_t *dest)
{
        enum wrtd_status status;

        /* Be sure the HMQ length is known. */
        status = wrtd_fill_config(wrtd);
        WRTD_RETURN_IF_ERROR(status);

        while (count > 0) {
                uint32_t tlen;
                struct trtl_msg msg;
                int err;
                struct wrtd_io_msg *io_msg;

                /* Partial transaction length.  */
                tlen = wrtd->hmq_words[cpu] - sizeof(struct wrtd_io_msg) / 4;
                if (tlen > count)
                        tlen = count;

                memset(&msg.hdr, 0, sizeof(struct trtl_hmq_header));
                msg.hdr.flags = TRTL_HMQ_HEADER_FLAG_RPC;
                msg.hdr.msg_id = WRTD_ACTION_READW;
                msg.hdr.len = sizeof(struct wrtd_io_msg) / 4;

                io_msg = (struct wrtd_io_msg *)msg.data;
                io_msg->addr = addr;
                io_msg->nwords = tlen;

                err = trtl_msg_sync(wrtd->trtl, cpu, 0, &msg, &msg,
                                    WRTD_DEFAULT_TIMEOUT);

                if (err)
                        return wrtd_return_error
                                (wrtd, WRTD_ERROR_UNEXPECTED_RESPONSE,
                                 "sync msg error(readw): %s",
                                 trtl_strerror(errno));

                if (msg.hdr.msg_id != WRTD_ACTION_READW
                    || msg.hdr.len != tlen)
                        return wrtd_return_error
                                (wrtd, WRTD_ERROR_UNEXPECTED_RESPONSE,
                                 "sync msg (readw): bad id 0x%x/len %d",
                                 msg.hdr.msg_id, msg.hdr.len);

                memcpy(dest, msg.data, tlen * sizeof(uint32_t));

                dest += tlen;
                count -= tlen;
        }
        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_msg_writew(struct wrtd_dev *wrtd, unsigned cpu,
                                 uint32_t addr, uint32_t count,
                                 const uint32_t *src)
{
        enum wrtd_status status;

        /* Be sure the HMQ length is known. */
        status = wrtd_fill_config(wrtd);
        WRTD_RETURN_IF_ERROR(status);

        while (count > 0) {
                uint32_t tlen;
                struct trtl_msg msg;
                int err;
                struct wrtd_io_msg *io_msg;

                /* Partial transaction length.  */
                tlen = wrtd->hmq_words[cpu] - sizeof(struct wrtd_io_msg) / 4;
                if (tlen > count)
                        tlen = count;

                memset(&msg.hdr, 0, sizeof(struct trtl_hmq_header));
                msg.hdr.flags = TRTL_HMQ_HEADER_FLAG_RPC;
                msg.hdr.msg_id = WRTD_ACTION_WRITEW;
                msg.hdr.len = (sizeof(struct wrtd_io_msg) / 4) + tlen;

                io_msg = (struct wrtd_io_msg *)msg.data;
                io_msg->addr = addr;
                io_msg->nwords = tlen;
                memcpy(io_msg->data, src, tlen * sizeof(uint32_t));

                err = trtl_msg_sync(wrtd->trtl, cpu, 0, &msg, &msg,
                                    WRTD_DEFAULT_TIMEOUT);

                if (err)
                        return wrtd_return_error
                                (wrtd, WRTD_ERROR_UNEXPECTED_RESPONSE,
                                 "sync msg error(writew): %s",
                                 trtl_strerror(errno));

                if (msg.hdr.msg_id != WRTD_ACTION_WRITEW
                    || msg.hdr.len != 0)
                        return wrtd_return_error
                                (wrtd, WRTD_ERROR_UNEXPECTED_RESPONSE,
                                 "sync msg (writew): bad id 0x%x/len %d",
                                 msg.hdr.msg_id, msg.hdr.len);

                src += tlen;
                count -= tlen;
        }
        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_write_root_flags(struct wrtd_dev *wrtd, unsigned cpu)
{
        enum wrtd_status status;

        status = wrtd_msg_writew
                (wrtd, cpu,
                 wrtd->root_addr[cpu] + offsetof(struct wrtd_root, log_flags),
                 1, (uint32_t *)&wrtd->roots[cpu].log_flags);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_validate_id(struct wrtd_dev *wrtd, const char *rep_cap_id)
{
        if (strlen(rep_cap_id) > WRTD_ID_LEN)
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_BAD_REP_CAP_ID,
                         "repeated capability name is too long");
        if (wrtd_id_null(rep_cap_id))
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_BAD_REP_CAP_ID,
                         "repeated capability name is null");

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_add_alarm(struct wrtd_dev *wrtd,
                                const char *rep_cap_id)
{
        enum wrtd_status status;
        unsigned int i;
        int idx;
        struct wrtd_alarm *alarm;

        /* 1. Validate rep_cap_id.  */
        status = wrtd_validate_id(wrtd, rep_cap_id);
        WRTD_RETURN_IF_ERROR(status);

        /* 2. Load all alarms.  */
        status = wrtd_fill_alarms(wrtd);
        WRTD_RETURN_IF_ERROR(status);

        /* 3. Find a free slot, check not already used.  */
        idx = -1;
        for (i = 0; i < wrtd->nbr_alarms; i++) {
                if (wrtd_id_eq(wrtd->alarms[i].alarm.event.id, rep_cap_id)) {
                        return wrtd_return_error(wrtd, WRTD_ERROR_ALARM_EXISTS,
                                                 "alarm already exists");
                }
                if (idx < 0 && wrtd_id_null(wrtd->alarms[i].alarm.event.id))
                        idx = i;
        }
        if (idx == -1)
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_ALARM_OUT_OF_RESOURCES,
                         "no alarm available");

        /* 4. Initialize with default values.  */
        alarm = &wrtd->alarms[idx].alarm;
        /* That should be the default, but... */
        memset(alarm, 0, sizeof(struct wrtd_alarm));
        wrtd_id_copy(alarm->event.id, rep_cap_id);

        /* 5. Write-back.  */
        status = wrtd_write_alarm(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_find_alarm(struct wrtd_dev *wrtd,
                                 const char *rep_cap_id,
                                 unsigned *idx)
{
        enum wrtd_status status;
        unsigned i;

        /* 1. Validate rep_cap_id.  */
        status = wrtd_validate_id(wrtd, rep_cap_id);
        WRTD_RETURN_IF_ERROR(status);

        /* 2. Load all alarms.  */
        status = wrtd_fill_alarms(wrtd);
        WRTD_RETURN_IF_ERROR(status);

        /* 3. Search it.  */
        for (i = 0; i < wrtd->nbr_alarms; i++)
                if (wrtd_id_eq(wrtd->alarms[i].alarm.event.id, rep_cap_id)) {
                        *idx = i;
                        return WRTD_SUCCESS;
                }

        return wrtd_return_error(wrtd, WRTD_ERROR_ALARM_DOES_NOT_EXIST,
                                 "alarm does not exist");
}

enum wrtd_status wrtd_remove_alarm(struct wrtd_dev *wrtd,
                                   const char *rep_cap_id)
{
        enum wrtd_status status;
        unsigned idx;
        struct wrtd_alarm *alarm;

        status = wrtd_find_alarm(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        /* Clear it.  */
        alarm = &wrtd->alarms[idx].alarm;
        memset(alarm, 0, sizeof(struct wrtd_alarm));

        /* Write-back.  */
        status = wrtd_write_alarm(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_id_copy_buf(struct wrtd_dev *wrtd,
                                  char *dst, int32_t buf_size, const char *id)
{
        if(buf_size < WRTD_ID_LEN)
                return wrtd_return_error(wrtd, WRTD_ERROR_BUFFER_TOO_SHORT,
                                         "buffer to short for an id");
        memset(dst, 0, WRTD_ID_LEN);
        wrtd_id_copy(dst, id);
        if(buf_size > WRTD_ID_LEN)
                dst[WRTD_ID_LEN] = 0;
        return WRTD_SUCCESS;
}


enum wrtd_status wrtd_get_alarm_id(struct wrtd_dev *wrtd,
                                   int32_t index,
                                   int32_t name_buffer_size,
                                   char *rep_cap_id)
{
        enum wrtd_status status;
        unsigned i;
        unsigned n;

        /* Load all alarms.  */
        status = wrtd_fill_alarms(wrtd);
        WRTD_RETURN_IF_ERROR(status);

        /* 3. Search it.  */
        for (n = 0, i = 0; i < wrtd->nbr_alarms; i++) {
                const char *aid = wrtd->alarms[i].alarm.event.id;
                if (wrtd_id_null(aid))
                        continue;
                if (n == index) {
                        status = wrtd_id_copy_buf
                                (wrtd, rep_cap_id, name_buffer_size, aid);
                        WRTD_RETURN_IF_ERROR(status);
                        return WRTD_SUCCESS;
                }
                n++;
        }

        return wrtd_return_error(wrtd, WRTD_ERROR_INVALID_VALUE,
                                 "invalid alarm index");
}

enum wrtd_status wrtd_log_read(struct wrtd_dev *wrtd,
                               struct wrtd_log_entry *log,
                               int poll_timeout)
{
        enum wrtd_status status;
	struct trtl_msg msg;
	struct polltrtl p[WRTD_MAX_CPUS];
	int ret, i;

        /* To get nbr cpus.  */
        status = wrtd_fill_config(wrtd);
        WRTD_RETURN_IF_ERROR(status);


	for (i = 0; i < wrtd->nbr_cpus; ++i) {
		p[i].trtl = wrtd->trtl;
		p[i].idx_hmq = WRTD_HMQ;
                p[i].idx_cpu = i;
		p[i].events = POLLIN;
	}

	/* Clean up errno to be able to distinguish between error cases and
	   normal behaviour when the function return less messages
	   than expected */
        errno = 0;

        ret = trtl_msg_poll(p, wrtd->nbr_cpus, poll_timeout);
        if (ret == 0)
                return wrtd_return_error(wrtd, WRTD_ERROR_TIMEOUT,
                                         "wrtd_log_read timeout");
        if (ret < 0)
                return wrtd_return_error (wrtd, WRTD_ERROR_UNEXPECTED_RESPONSE,
                                          "poll error(log_read): %s",
                                          trtl_strerror(errno));

        for (i = 0; i < wrtd->nbr_cpus; ++i) {
                if (!(p[i].revents & POLLIN))
                        continue;
                ret = trtl_msg_async_recv(wrtd->trtl, i, WRTD_HMQ, &msg, 1);
                if (ret <= 0)
                        return wrtd_return_error
                                (wrtd, WRTD_ERROR_UNEXPECTED_RESPONSE,
                                 "async msg error(log_read): %s",
                                 trtl_strerror(errno));
                memcpy(log, msg.data, sizeof(struct wrtd_log_entry));
                return WRTD_SUCCESS;
        }
        /* Should not happen.  */
        return wrtd_return_error(wrtd, WRTD_ERROR_TIMEOUT,
                                 "wrtd_log_read timeout");
}

enum wrtd_status wrtd_alloc_rules(struct wrtd_dev *wrtd,
                                  struct wrtd_lib_rule **rulesp)
{
        struct wrtd_lib_rule *rules;
        unsigned n;
        unsigned cpu;

        rules = (struct wrtd_lib_rule *)
                calloc(wrtd->nbr_rules, sizeof(struct wrtd_lib_rule));
        if (rules == NULL)
                return wrtd_return_error(wrtd, WRTD_ERROR_OUT_OF_MEMORY,
                                         "cannot allocate %u rules",
                                         wrtd->nbr_rules);

        n = 0;
        for (cpu = 0; cpu < wrtd->nbr_cpus; cpu++) {
                unsigned nbr_rules = wrtd->roots[cpu].nbr_rules;
                unsigned idx;
                for (idx = 0; idx < nbr_rules; idx++) {
                        rules[n].cpu = cpu;
                        rules[n].local_idx = idx;
                        n++;
                }
        }
        assert(n == wrtd->nbr_rules);
        *rulesp = rules;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_fill_rules(struct wrtd_dev *wrtd)
{
        enum wrtd_status status;
        unsigned i;
        unsigned cpu;

        if (wrtd->rules != NULL)
                return WRTD_SUCCESS;

        status = wrtd_fill_roots(wrtd);
        WRTD_RETURN_IF_ERROR(status);

        /* Count number of rules.  */
        assert(wrtd->nbr_rules == 0);
        for (cpu = 0; cpu < wrtd->nbr_cpus; cpu++)
                wrtd->nbr_rules += wrtd->roots[cpu].nbr_rules;

        /* Allocates.  */
        status = wrtd_alloc_rules(wrtd, &wrtd->rules);
        WRTD_RETURN_IF_ERROR(status);

        /* Read rules from cpus.  */
        for (i = 0; i < wrtd->nbr_rules; i++) {
                unsigned cpu = wrtd->rules[i].cpu;
                unsigned int addr;

                addr = (wrtd->roots[cpu].rules_addr
                        + wrtd->rules[i].local_idx * sizeof(struct wrtd_rule));
                status = wrtd_msg_readw
                        (wrtd, cpu, addr, sizeof(struct wrtd_rule) / 4,
                         (uint32_t *)&wrtd->rules[i].rule);
                if (status != WRTD_SUCCESS) {
                        free(wrtd->rules);
                        wrtd->rules = NULL;
                        return status;
                }
        }

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_write_rule(struct wrtd_dev *wrtd, unsigned idx)
{
        enum wrtd_status status;
        struct wrtd_lib_rule *rule = &wrtd->rules[idx];

        status = wrtd_msg_writew
                (wrtd, rule->cpu,
                 (wrtd->roots[rule->cpu].rules_addr
                  + rule->local_idx * sizeof(struct wrtd_rule)),
                 sizeof(struct wrtd_rule) / 4,
                 (uint32_t *)&rule->rule);

        return status;
}

enum wrtd_status wrtd_add_rule(struct wrtd_dev *wrtd,
                                const char *rep_cap_id)
{
        enum wrtd_status status;
        unsigned int i;
        int idx;
        struct wrtd_rule *rule;

        /* 1. Validate rep_cap_id.  */
        status = wrtd_validate_id(wrtd, rep_cap_id);
        WRTD_RETURN_IF_ERROR(status);

        /* 2. Load all rules.  */
        status = wrtd_fill_rules(wrtd);
        WRTD_RETURN_IF_ERROR(status);

        /* 3. Find a free slot, check not already used.  */
        idx = -1;
        for (i = 0; i < wrtd->nbr_rules; i++) {
                if (wrtd_id_eq(wrtd->rules[i].rule.conf.id, rep_cap_id)) {
                        return wrtd_return_error(wrtd, WRTD_ERROR_RULE_EXISTS,
                                                 "rule already exists");
                }
                if (idx < 0
                    && wrtd_id_null(wrtd->rules[i].rule.conf.id))
                        idx = i;
        }
        if (idx == -1)
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_RULE_OUT_OF_RESOURCES,
                         "no rule available");

        /* 4. Initialize with default values.  */
        rule = &wrtd->rules[idx].rule;
        /* That should be the default, but... */
        memset(rule, 0, sizeof(struct wrtd_rule));
        wrtd_id_copy(rule->conf.id, rep_cap_id);
        rule->conf.send_late = true;

        /* 5. Write-back.  */
        status = wrtd_write_rule(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_find_rule(struct wrtd_dev *wrtd,
                                const char *rep_cap_id,
                                unsigned *idx)
{
        enum wrtd_status status;
        unsigned i;

        /* 1. Validate rep_cap_id.  */
        status = wrtd_validate_id(wrtd, rep_cap_id);
        WRTD_RETURN_IF_ERROR(status);

        /* 2. Load all rules.  */
        status = wrtd_fill_rules(wrtd);
        WRTD_RETURN_IF_ERROR(status);

        /* 3. Search it.  */
        for (i = 0; i < wrtd->nbr_rules; i++)
                if (wrtd_id_eq(wrtd->rules[i].rule.conf.id, rep_cap_id)) {
                        *idx = i;
                        return WRTD_SUCCESS;
                }

        return wrtd_return_error(wrtd, WRTD_ERROR_RULE_DOES_NOT_EXIST,
                                 "rule does not exist");
}

enum wrtd_status wrtd_get_rule_id(struct wrtd_dev *wrtd,
                                  int32_t index,
                                  int32_t name_buffer_size,
                                  char *rep_cap_id)
{
        enum wrtd_status status;
        unsigned i;
        unsigned n;

        /* Load all rules.  */
        status = wrtd_fill_rules(wrtd);
        WRTD_RETURN_IF_ERROR(status);

        /* 3. Search it.  */
        for (n = 0, i = 0; i < wrtd->nbr_rules; i++) {
                const char *aid = wrtd->rules[i].rule.conf.id;
                if (wrtd_id_null(aid))
                        continue;
                if (n == index) {
                        status = wrtd_id_copy_buf
                                (wrtd, rep_cap_id, name_buffer_size, aid);
                        WRTD_RETURN_IF_ERROR(status);
                        return WRTD_SUCCESS;
                }
                n++;
        }

        return wrtd_return_error(wrtd, WRTD_ERROR_INVALID_VALUE,
                                 "invalid rule index");
}

enum wrtd_status wrtd_remove_rule(struct wrtd_dev *wrtd,
                                  const char *rep_cap_id)
{
        enum wrtd_status status;
        unsigned idx;
        struct wrtd_rule *rule;

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx);
        WRTD_RETURN_IF_ERROR(status);

        rule = &wrtd->rules[idx].rule;
        if (rule->conf.enabled)
                return wrtd_return_error(wrtd, WRTD_ERROR_RULE_ENABLED,
                                         "rule is enabled");

        /* Clear it.  */
        memset(rule, 0, sizeof(struct wrtd_rule));

        /* Write-back.  */
        status = wrtd_write_rule(wrtd, idx);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}
