#ifndef __LIBWRTD_PRIVATE__H__
#define __LIBWRTD_PRIVATE__H__

#include <strings.h>
#include "libwrtd.h"
#include "libmockturtle.h"

/* Maximum number of CPUS per MT (allow less than the MT arch).  */
#define WRTD_MAX_CPUS 4

/* Maximum number of devices per CPU. */
#define WRTD_MAX_DEVS 4

#define WRTD_RETURN_IF_ERROR(status) if(status != WRTD_SUCCESS) return status

/* Alarm when in user space.  */
struct wrtd_lib_alarm {
        /* CPU on which the alarm is.  */
        unsigned int cpu;
        /* Alarm index for the cpu.  */
        unsigned int idx;
        /* Alarm data.  */
        struct wrtd_alarm alarm;
};

struct wrtd_lib_rule {
        /* MT CPU on which the rule is placed.  */
        unsigned int cpu;
        /* Rule index on the local MT cpu.  */
        unsigned int local_idx;
        /* Rule.  */
        struct wrtd_rule rule;
};

struct wrtd_dev {
        /* Associated mock-turtle device.  */
        struct trtl_dev *trtl;

        /* Device/cpu configuration.  */
        const struct wrtd_mt_config *config;

        /* Last error.  */
        enum wrtd_status err;

        /* Long error message.  */
        char error_msg[256];

        /* Number of cpus. Valid if > 0. */
        uint32_t nbr_cpus;

        /* Per-cpu hmq length (from MT config) - this is needed for
           readw/writew. */
        uint32_t hmq_words[WRTD_MAX_CPUS];

        /* Per-cpu root address.  */
        uint32_t root_addr[WRTD_MAX_CPUS];

        /* Roots.
           Valid only when versions are not 0.  */
        struct wrtd_root roots[WRTD_MAX_CPUS];

        /* Rules.  */
        uint32_t nbr_rules;
        struct wrtd_lib_rule *rules;

        /* Alarms.  */
        uint32_t nbr_alarms;
        struct wrtd_lib_alarm *alarms;
};

/* Save the error in DEV and return the error code.  */
extern enum wrtd_status wrtd_return_error(struct wrtd_dev *dev,
                                          enum wrtd_status error_code,
                                          const char *format, ...)
        __attribute__((format(printf, 3, 4)));


/* Get current error message.  Simpler than the official API.  */
extern const char *wrtd_get_error_msg(struct wrtd_dev *dev);

/* Get the list of firmware expected by cpus.
   IDS must be at least WRTD_MAX_CPUS long.  */
extern enum wrtd_status wrtd_get_firmware_id(struct wrtd_dev *dev,
                                             unsigned *nbr_ids,
                                             uint32_t *ids);

/* For wrtd_load_firmware.  */
struct wrtd_firmware {
        const char *path;
        uint32_t id;
};

enum wrtd_status wrtd_load_firmware(struct wrtd_dev *dev, unsigned verbose,
                                    unsigned int nbr_firmwares,
                                    struct wrtd_firmware *fw);

enum wrtd_dir { wrtd_input, wrtd_output };

/* Device/FMC description. */
struct wrtd_mt_device {
        /* Name of the FMC. */
        const char *name;
        /* Direction.  */
        enum wrtd_dir dir;
        /* Number of channels.  */
        unsigned int nbr_chs;
        /* Channels name.  */
        const char * const *channels;
        /* HW configuration.  */
        enum wrtd_status (*configure)(struct wrtd_dev *wrtd,
                                      unsigned cpu, unsigned dev,
                                      unsigned int mask);
};

/* Per CPU configuration.  */
struct wrtd_mt_cpu_config {
        /* Firmware id to be loaded. Matches the firwmare app id.  */
        uint32_t fw_id;
        /* Devices/FMC controlled by this cpu. */
        unsigned nbr_devices;
        const struct wrtd_mt_device *devices[WRTD_MAX_DEVS];
        /* TODO: add network in, network out, loopback in.  */
};

struct wrtd_mt_config {
        /* From MT config rom.  */
        uint32_t mt_app_id;
        uint32_t nbr_cpus;
        /* CPU for network rx/tx.  */
        uint32_t tx_cpu;
        uint32_t rx_cpu;
        /* APP_ID for each cpu.  */
        struct wrtd_mt_cpu_config cpus[WRTD_MAX_CPUS];
};

/* Find device configuration for wrtd.  */
enum wrtd_status wrtd_find_config(struct wrtd_dev *wrtd,
                                  const struct wrtd_mt_config **config);

/* Be sure the config is set in WRTD.  */
enum wrtd_status wrtd_fill_config(struct wrtd_dev *wrtd);

/* Do the get_config action. */
enum wrtd_status wrtd_msg_get_config(struct wrtd_dev *wrtd, unsigned cpu,
                                     struct wrtd_config_msg *res);

/* Read words from CPU space.  */
enum wrtd_status wrtd_msg_readw(struct wrtd_dev *wrtd, unsigned cpu,
                                uint32_t addr, uint32_t count,
                                uint32_t *dest);

/* Write words to CPU space.  */
enum wrtd_status wrtd_msg_writew(struct wrtd_dev *wrtd, unsigned cpu,
                                 uint32_t addr, uint32_t count,
                                 const uint32_t *src);

/* Check if rep_cap_id is valid (not null, length). */
enum wrtd_status wrtd_validate_id(struct wrtd_dev *wrtd,
                                  const char *rep_cap_id);

/* Ensure the roots have been fetched from the CPUs.  */
enum wrtd_status wrtd_fill_roots(struct wrtd_dev *wrtd);

enum wrtd_status wrtd_write_root_flags(struct wrtd_dev *wrtd, unsigned cpu);

/* Ensure the alarms have been fetched from the CPUs.  */
enum wrtd_status wrtd_fill_alarms(struct wrtd_dev *wrtd);

/* Write back alarm idx. */
enum wrtd_status wrtd_write_alarm(struct wrtd_dev *wrtd, unsigned idx);

/* Find alarm index by id. */
enum wrtd_status wrtd_find_alarm(struct wrtd_dev *wrtd,
                                 const char *rep_cap_id,
                                 unsigned *idx);

/* Find rule index by id. */
enum wrtd_status wrtd_find_rule(struct wrtd_dev *wrtd,
                                const char *rep_cap_id,
                                unsigned *idx);

/* Internal function: allocate rules and set cpu,idx.  */
enum wrtd_status wrtd_alloc_rules(struct wrtd_dev *wrtd,
                                  struct wrtd_lib_rule **rulesp);

/* Write-back rule. */
enum wrtd_status wrtd_write_rule(struct wrtd_dev *wrtd, unsigned idx);

/* Ensure the rules have been fetched from the CPUs.  */
enum wrtd_status wrtd_fill_rules(struct wrtd_dev *wrtd);

/* Count number of rules.  */
enum wrtd_status wrtd_attr_get_rule_count(struct wrtd_dev *wrtd,
                                          unsigned *value);

/* Reconfigure route tables and write-back rules.  */
enum wrtd_status wrtd_reconfigure(struct wrtd_dev *wrtd);

static inline bool wrtd_id_eq(const char *l, const char *r)
{
        return strncmp(l, r, WRTD_ID_LEN) == 0;
}

static inline bool wrtd_id_null(const char *id)
{
        return id[0] == 0;
}

static inline void wrtd_id_copy(char *dst, const char *src)
{
        strncpy(dst, src, WRTD_ID_LEN);
}

/* Copy ID to DST of length BUF_SIZE.  Check length, pad with null.  */
enum wrtd_status wrtd_id_copy_buf(struct wrtd_dev *wrtd,
                                  char *dst, int32_t buf_size, const char *id);

#endif /* __LIBWRTD_PRIVATE__H__ */
