#ifndef __WRTD_LIB_H__
#define __WRTD_LIB_H__
/** @file libwrtd-ng.h */

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <stdbool.h>
#include "wrtd-common.h"

// NOTE: corresponds roughly to IVI ViSession
struct wrtd_dev;

/**
 * @enum wrtd_status
 * White Rabbit Trigger Distribution warnings and errors
 */
enum wrtd_status {
	/* Codes inspired by IVI 3.2 */
	WRTD_SUCCESS = 0,
	WRTD_ERROR_BASE = 0xBFFA3000,
	WRTD_ERROR_CANNOT_RECOVER,
	WRTD_ERROR_INVALID_ATTRIBUTE,
	WRTD_ERROR_ATTR_NOT_WRITEABLE,
	WRTD_ERROR_ATTR_NOT_READABLE,
	WRTD_ERROR_INVALID_VALUE,
	WRTD_ERROR_NOT_INITIALIZED,
	WRTD_ERROR_MISSING_OPTION_NAME,
	WRTD_ERROR_MISSING_OPTION_VALUE,
	WRTD_ERROR_BAD_OPTION_NAME,
	WRTD_ERROR_BAD_OPTION_VALUE,
	WRTD_ERROR_OUT_OF_MEMORY,
	WRTD_ERROR_OPERATION_PENDING,
	WRTD_ERROR_NULL_POINTER,
	WRTD_ERROR_UNEXPECTED_RESPONSE,
	WRTD_ERROR_RESET_FAILED,
	WRTD_ERROR_RESOURCE_UNKNOWN,

        /* Resource (alar, rule) is not disabled and cannot be changed.  */
        WRTD_ERROR_RESOURCE_ACTIVE,

        /* Require global rep cap.  */
        WRTD_ERROR_ATTR_INVALID_REP_CAP,

        /* Feature not implemented.  */
        WRTD_ERROR_NOT_IMPLEMENTED,

        /* Incorrect repeated capability id: too long, invalid character... */
        WRTD_ERROR_BAD_REP_CAP_ID,

        /* Timeout while reading log.  */
        WRTD_ERROR_TIMEOUT,

        /* Output buffer is not long enough.  */
        WRTD_ERROR_BUFFER_TOO_SHORT,

	WRTD_ERROR_UNKNOWN_NAME_IN_SELECTOR, // NOTE: selector == rep_cap_id
	/* Codes inspired by IviLxiSync 3.15 */
	WRTD_ERROR_ALARM_TIME_INVALID,
	WRTD_ERROR_ALARM_EXISTS,
	WRTD_ERROR_ALARM_DOES_NOT_EXIST,
	WRTD_ERROR_ALARM_OUT_OF_RESOURCES,
	WRTD_ERROR_RULE_EXISTS,
	WRTD_ERROR_RULE_DOES_NOT_EXIST,
	WRTD_ERROR_RULE_OUT_OF_RESOURCES,
	WRTD_ERROR_RULE_INVALID,
        WRTD_ERROR_RULE_ENABLED,
	WRTD_ERROR_CANT_REMOVE_RESERVED_REP_CAP,
	__WRTD_ERROR_MAX_NUMBER,
};

/**
 * @enum wrtd_attr
 * White Rabbit Trigger Distribution attribute ID definions
 */
enum wrtd_attr {
	__WRTD_ATTR_BASE = 950000,
	WRTD_MAJOR_VERSION, // RO, int32
	WRTD_MINOR_VERSION, // RO, int32
	WRTD_ATTR_EVENT_LOG_ENTRY_COUNT, // RO, int32
	WRTD_ATTR_EVENT_LOG_ENABLED, //RW, bool
	// TODO: add log levels/masks here
	WRTD_ATTR_IS_TIME_MASTER, // RO, bool
	WRTD_ATTR_IS_TIME_SYNCHRONIZED, // RO, bool

        // Number of alarms (global attribute).
	WRTD_ATTR_ALARM_COUNT, //RO, int32

        // Enable/disable an alarm.
	WRTD_ATTR_ALARM_ENABLED, //RW, bool

	WRTD_ATTR_ALARM_SETUP_TIME, //RW, tstamp
	WRTD_ATTR_ALARM_TIME, //RW, tstamp
	// Specifies the alarm period. 0 means no repetitions.
	WRTD_ATTR_ALARM_PERIOD, //RW, tstamp
	// Specifies the number of times the alarm will occur at the period specified by
	// WRTD_ATTR_ALARM_PERIOD. 0 means infinite. 1 means exactly one alarm will occur.
	WRTD_ATTR_ALARM_REPEAT_COUNT, //RW, int32

        // Number of rules (global attribute).
	WRTD_ATTR_RULE_COUNT, //RO, int32

        // True if the rule is enabled (ie active).
	WRTD_ATTR_RULE_ENABLED, //RW, bool
        // Number of time the rule will apply.  0 means infinite.  When read,
        // return the remaining number.
	WRTD_ATTR_RULE_REPEAT_COUNT, //RW, int32
        // Source of the event.
	WRTD_ATTR_RULE_SOURCE, //RW, string
        // Destination of the event.
	WRTD_ATTR_RULE_DESTINATION, //RW, string
        // If true, events from the past are still handled; otherwise they
        // are discarded.
	WRTD_ATTR_RULE_SEND_LATE, //RW, bool
        // Delay the event.
	WRTD_ATTR_RULE_DELAY, //RW, tstamp
        // Discard new events between the last one to this value.
	WRTD_ATTR_RULE_HOLDOFF, //RW, tstamp
        // Realign the timestamp to a multiple of period and then add
        // factor * period. This is done after the delay.
	WRTD_ATTR_RULE_RESYNC_PERIOD, //RW, tstamp
	WRTD_ATTR_RULE_RESYNC_FACTOR, //RW, int32

	// TODO: add __ boundaries between groups to make it easier
	// to do input validation
	WRTD_ATTR_STAT_RULE_RX_EVENTS, //RO, int32
	WRTD_ATTR_STAT_RULE_RX_LAST, //RO, tstamp
	WRTD_ATTR_STAT_RULE_TX_EVENTS, //RO, int32
	WRTD_ATTR_STAT_RULE_TX_LAST, //RO, tstamp
	WRTD_ATTR_STAT_RULE_MISSED_EVENTS_LATE, //RO, int32
	WRTD_ATTR_STAT_RULE_MISSED_EVENTS_HOLDOFF, //RO, int32
	WRTD_ATTR_STAT_RULE_MISSED_EVENTS_OVERFLOW, //RO, int32
	WRTD_ATTR_STAT_RULE_MISSED_LAST, //RO, tstamp

        /* Latency and period (excluding events discarded due to holdoff).  */
	WRTD_ATTR_STAT_RULE_RX_LATENCY_MIN, //RO, tstamp (maybe ns only valid)
	WRTD_ATTR_STAT_RULE_RX_LATENCY_MAX, //RO, tstamp (maybe ns only valid)
	WRTD_ATTR_STAT_RULE_RX_LATENCY_AVG, //RO, tstamp (maybe ns only valid)
	WRTD_ATTR_STAT_RULE_RX_PERIOD_MIN, //RO, tstamp (maybe ns only valid)
	WRTD_ATTR_STAT_RULE_RX_PERIOD_MAX, //RO, tstamp (maybe ns only valid)
	WRTD_ATTR_STAT_RULE_RX_PERIOD_AVG, //RO, tstamp (maybe ns only valid)
	__WRTD_ATTR_MAX_NUMBER,
};

/**
 * A repeated capability identifier for global attributes
 */
#define WRTD_GLOBAL_REP_CAP_ID "WGRCI"

/**
 * @enum wrtd_log_type
 * White Rabbit Trigger Distribution log entry type
 */
enum wrtd_log_type {
        // TODO: add log types
	__WRTD_MAX_LOG_TYPE_NUMBER,
};

#if 0
/**
 * WRTD log event descriptor
 */
struct wrtd_log_entry {
        // TODO: verify fields
        enum  wrtd_log_type type; /**< type of logging */
	uint32_t seq; /**< log sequence number */
	char *id; /**< trigger id associated with the log event */
	struct wrtd_tstamp ts; /**< when the log message was sent from
                                  the RT application*/
};
#endif

/**
 * @file libwrtd-base.c
 */
/**
 * @defgroup base
 * Set of functions to manage the basic device and library configuration.
 * @{
 */
// Ideally, also be able to list possible names
extern enum wrtd_status wrtd_init(const char *resource_name,
				  bool reset,
				  const char *options_str,
				  struct wrtd_dev **dev);
extern enum wrtd_status wrtd_close(struct wrtd_dev *dev);
extern enum wrtd_status wrtd_reset(struct wrtd_dev *dev);
extern enum wrtd_status wrtd_get_error(struct wrtd_dev *dev,
                                       enum wrtd_status *error_code,
                                       uint32_t buffer_size,
                                       char error_description);
extern enum wrtd_status wrtd_error_message(struct wrtd_dev *dev,
					   enum wrtd_status err_code,
					   char *err_message);


/* Attributes.  */
extern enum wrtd_status wrtd_get_attr_int32(struct wrtd_dev *dev,
					    const char *rep_cap_id,
					    enum wrtd_attr id,
					    int32_t *value);
extern enum wrtd_status wrtd_set_attr_int32(struct wrtd_dev *dev,
					    const char *rep_cap_id,
					    enum wrtd_attr id,
					    int32_t value);
extern enum wrtd_status wrtd_set_attr_int64(struct wrtd_dev *dev,
					    const char *rep_cap_id,
					    enum wrtd_attr id,
					    int64_t value);
extern enum wrtd_status wrtd_get_attr_bool(struct wrtd_dev *dev,
					   const char *rep_cap_id,
					   enum wrtd_attr id,
					   bool *value);
extern enum wrtd_status wrtd_set_attr_bool(struct wrtd_dev *dev,
					   const char *rep_cap_id,
					   enum wrtd_attr id,
					   bool value);
extern enum wrtd_status wrtd_get_attr_tstamp(struct wrtd_dev *dev,
					     const char *rep_cap_id,
					     enum wrtd_attr id,
					     struct wrtd_tstamp *value);
extern enum wrtd_status wrtd_set_attr_tstamp(struct wrtd_dev *dev,
					     const char *rep_cap_id,
					     enum wrtd_attr id,
					     const struct wrtd_tstamp *value);
extern enum wrtd_status wrtd_set_attr_string(struct wrtd_dev *dev,
					     const char *rep_cap_id,
					     enum wrtd_attr id,
					     const char *value);
extern enum wrtd_status wrtd_get_attr_int64(struct wrtd_dev *dev,
					    const char *rep_cap_id,
					    enum wrtd_attr id,
					    int64_t *value);
extern enum wrtd_status wrtd_get_attr_string(struct wrtd_dev *dev,
					     const char *rep_cap_id,
					     enum wrtd_attr id,
					     int32_t value_buf_size,
					     char *value);

/**@}*/

/**
 * @file libwrtd-time.c
 */
/**
 * @defgroup time
 * Set of functions to manage time
 * @{
 */
extern enum wrtd_status wrtd_get_sys_time(struct wrtd_dev *dev,
					  struct wrtd_tstamp *time);

/**@}*/

/**
 * @file libwrtd-log.c
 */
/**
 * @defgroup logging
 * Set of logging functions
 * @{
 */
extern enum wrtd_status wrtd_log_read(struct wrtd_dev *dev,
				      struct wrtd_log_entry *log,
				      int poll_timeout);
// Maybe also add non-blocking wrtd_get_next_log_entry()
extern enum wrtd_status wrtd_clear_log(struct wrtd_dev *dev);
/**@}*/

/**
 * @file libwrtd-alarm.c
 */
/**
 * @defgroup alarms
 * Set of functions to configure alarms as
 * sources for the generation of events
 * @{
 */
extern enum wrtd_status wrtd_add_alarm(struct wrtd_dev *dev,
				       const char *rep_cap_id);
extern enum wrtd_status wrtd_disable_all_alarms(struct wrtd_dev *dev);
extern enum wrtd_status wrtd_remove_alarm(struct wrtd_dev *dev,
					  const char *rep_cap_id);
extern enum wrtd_status wrtd_remove_all_alarms(struct wrtd_dev *dev);
extern enum wrtd_status wrtd_get_alarm_id(struct wrtd_dev *dev,
					  int32_t index,
					  int32_t name_buffer_size,
					  char *rep_cap_id);

/**
 * @file libwrtd-rule.c
 */
/**
 * @defgroup rules
 * Set of functions to trigger on and
 * schedule the generation of events
 * @{
 */
// rule sources can be a) reserved event ids, b) alarm* c) any other string
// which will be interpreted as a net msg event id
extern enum wrtd_status wrtd_add_rule(struct wrtd_dev *dev,
				      const char *rep_cap_id);
extern enum wrtd_status wrtd_disable_all_rules(struct wrtd_dev *dev);
extern enum wrtd_status wrtd_remove_rule(struct wrtd_dev *dev,
					 const char *rep_cap_id);
extern enum wrtd_status wrtd_remove_all_rules(struct wrtd_dev *dev);
extern enum wrtd_status wrtd_get_rule_id(struct wrtd_dev *dev,
					 int32_t index,
					 int32_t name_buffer_size,
					 char *rep_cap_id);
extern enum wrtd_status wrtd_reset_rule_stats(struct wrtd_dev *dev,
					      const char *rep_cap_id);

/**@}*/

// ===========================================================================
// BEGIN DEPRECATED
// ===========================================================================

/**
 * Status description for an input channel
 */
struct wrtd_input_state {
	struct wrtd_tstamp tdc_timebase_offset; /**< TDC time base offset */
};

/**
 * Status description for an output channel
 */
struct wrtd_output_state {
	struct wrtd_tstamp pulse_width; /**< Pulse width */
};

/**@}*/


/**@}*/

// ===========================================================================
// END DEPRECATED
// ===========================================================================

#ifdef __cplusplus
};
#endif

#endif
