/*
 * Copyright (C) 2014-2016 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <getopt.h>
#include <inttypes.h>
#include "libwrtd.h"
#include "libwrtd-private.h"
#include "wrtd-internal.h"

static void help(void)
{
	fprintf(stderr, "wrtd-logging -D <device>\n");
	fprintf(stderr, "It shows logging information coming from Real-Time applications\n");
	fprintf(stderr, "-D device id\n");
	fprintf(stderr, "-n number of messages to read (0 means infinite)\n");
	exit(1);
}


static void print_logging (struct wrtd_dev *wrtd, int n_read)
{
	struct wrtd_log_entry log;
        enum wrtd_status status;
        unsigned int i;

	for(i = 0; i < n_read || n_read == 0; i++) {
                status = wrtd_log_read(wrtd, &log, -1);
                if (status != WRTD_SUCCESS) {
                        fprintf(stderr,
                                "Error: %s\n", wrtd_get_error_msg(wrtd));
                        return;
                }

                if (log.ts.seconds != 0 || log.ts.ns != 0) {
                        printf("At ");
                        print_ts(&log.event.ts);
                        putchar('\n');
                }

                switch(log.type) {
                case WRTD_LOG_MSG_EV_GENERATED:
                        printf("generated ");
			if (log.reason == WRTD_LOG_GENERATED_ALARM)
                                printf("alrm     ");
                        else if (log.reason >= WRTD_LOG_GENERATED_DEVICE_0
                                 && log.reason < WRTD_LOG_GENERATED_DEVICE_1)
                                printf("dev0,ch%u",
                                       log.reason - WRTD_LOG_GENERATED_DEVICE_0);
                        else if (log.reason >= WRTD_LOG_GENERATED_DEVICE_1
                                 && log.reason < WRTD_LOG_GENERATED_DEVICE_2)
                                printf("dev1,ch%u",
                                       log.reason - WRTD_LOG_GENERATED_DEVICE_1);
                        else
				printf("%-8u ", log.reason);
                        break;
                case WRTD_LOG_MSG_EV_CONSUMED:
                        printf("consumed  ");
                        switch(log.reason) {
                        case WRTD_LOG_CONSUMED_START:
                                printf("start   ");
                                break;
                        case WRTD_LOG_CONSUMED_DONE:
                                printf("done    ");
                                break;
                        default:
                                printf("%-7u ", log.reason);
                                break;
                        }
                        break;
                case WRTD_LOG_MSG_EV_DISCARDED:
                        printf("discarded ");
                        switch(log.reason) {
			case WRTD_LOG_DISCARD_NO_SYNC:
				printf("nosync  ");
				break;
			case WRTD_LOG_DISCARD_TIMEOUT:
				printf("timeout ");
				break;
			case WRTD_LOG_DISCARD_OVERFLOW:
				printf("ovrflow ");
				break;
                        default:
                                printf("%-7u ", log.reason);
                                break;
			}
                        break;
                case WRTD_LOG_MSG_EV_NETWORK:
                        printf("network   ");
                        switch(log.reason) {
                        case WRTD_LOG_NETWORK_TX:
                                printf("tx      ");
                                break;
                        case WRTD_LOG_NETWORK_RX:
                                printf("rx      ");
                                break;
                        default:
                                printf("%-7u ", log.reason);
                                break;
                        }
                        break;
                case WRTD_LOG_MSG_LINK:
                        printf("link      ");
                        break;
                default:
                        printf("?%d", log.type);
                        break;
                }
                printf(" id: %-8.*s", WRTD_ID_LEN, log.event.id);
                printf(" ev ts: ");
                print_ts(&log.event.ts);
                printf("\n");

                fflush(stdout);
	}
}

int main(int argc, char *argv[])
{
	struct wrtd_dev *wrtd;
        const char *dev_name = NULL;
        enum wrtd_status status;
        char *endptr;
	int n_read = 0;
	char c;

	while ((c = getopt (argc, argv, "hD:n:")) != -1) {
		switch (c) {
		default:
			help();
			break;
		case 'D':
                        dev_name = optarg;
			break;
		case 'n':
			n_read = strtoul(optarg, &endptr, 0);
			if (*endptr != 0) {
				fprintf(stderr, "bad value for -n");
				exit (1);
			}
			break;
		}
	}

	if (dev_name == NULL) {
		help();
		exit(1);
	}

        /* Open.  */
	status = wrtd_init(dev_name, 0, NULL, &wrtd);
	if (status != WRTD_SUCCESS) {
		fprintf(stderr, "Cannot open WRTD: %s\n",
                        wrtd_get_error_msg(wrtd));
                return 1;
	}

	print_logging (wrtd, n_read);

	wrtd_close(wrtd);
	exit(0);
}
